import os
import re
import numpy as np 
from dataclasses import MISSING, dataclass, field
from typing import Any, Dict, Iterable, List, Optional
from omegaconf import OmegaConf,DictConfig
from torchkeras.tabular.models.common import heads

def get_inferred_config(ds):
    categorical_dim = len(ds.categorical_cols)
    continuous_dim = len(ds.continuous_cols)
    if ds.task == "regression":
        output_dim = len(ds.target) if ds.target else None
    elif ds.task == "binary":
        output_dim = 1
        if ds.target:
            target_values = set(np.unique(ds.data[ds.target[0]].astype(int)))
            for value in target_values:
                assert  value in {0,1}, f'not valid targets: {target_values}'
                
    elif ds.task in ("multiclass","classification"):
        output_dim = len(np.unique(ds.data[ds.target[0]])) if ds.target else None
    else:
        raise ValueError(f"{ds.task} is an unsupported task.")
    
    categorical_cardinality = [
        int(x) + 1 for x in ds.data[ds.categorical_cols].max().values
    ]
    embedding_dims = [(x, min(50, (x + 1) // 2)) for x in categorical_cardinality]
        
    return InferredConfig(
        categorical_cols = ds.categorical_cols,
        continuous_cols = ds.continuous_cols,
        categorical_dim=categorical_dim,
        continuous_dim=continuous_dim,
        output_dim=output_dim,
        categorical_cardinality=categorical_cardinality,
        embedding_dims=embedding_dims
    )
    
def safe_merge_config(config: DictConfig, inferred_config: DictConfig) -> DictConfig:
    """Merge two configurations.

    Args:
        config: The base configuration.
        inferred_config: The custom configuration.

    Returns:
        The merged configuration.

    """
    # using base config values if exist
    inferred_config.embedding_dims = config.embedding_dims or inferred_config.embedding_dims
    merged_config = OmegaConf.merge(OmegaConf.to_container(config), OmegaConf.to_container(inferred_config))
    return merged_config

@dataclass
class InferredConfig:
    """Configuration inferred from the dataset.

    Args:
        categorical_cols (List[str]): categorical feature names
        
        continuous_cols (List[str]): continuous feature names
        
        categorical_dim (int): The number of categorical features

        continuous_dim (int): The number of continuous features

        output_dim (Optional[int]): The number of output targets

        categorical_cardinality (Optional[List[int]]): The number of unique values in categorical features

        embedding_dims (Optional[List]): The dimensions of the embedding for each categorical column as a
                list of tuples (cardinality, embedding_dim).

        embedded_cat_dim (int): The number of features or dimensions of the embedded categorical features

    """
    categorical_cols: List[str] = field(
        metadata={"help": "The names of categorical features"},
    )

    continuous_cols: List[str] = field(
        metadata={"help": "The names of continuous features"},
    )

    categorical_dim: int = field(
        metadata={"help": "The number of categorical features"},
    )
    
    continuous_dim: int = field(
        metadata={"help": "The number of continuous features"},
    )
    output_dim: Optional[int] = field(
        default=None,
        metadata={"help": "The number of output targets"},
    )
    categorical_cardinality: Optional[List[int]] = field(
        default=None,
        metadata={"help": "The number of unique values in categorical features"},
    )
    embedding_dims: Optional[List] = field(
        default=None,
        metadata={
            "help": "The dimensions of the embedding for each categorical column as a list of tuples "
            "(cardinality, embedding_dim)."
        },
    )
    embedded_cat_dim: int = field(
        init=False,
        metadata={"help": "The number of features or dimensions of the embedded categorical features"},
    )

    def __post_init__(self):
        if self.embedding_dims is not None:
            assert all(
                (isinstance(t, Iterable) and len(t) == 2) for t in self.embedding_dims
            ), "embedding_dims must be a list of tuples (cardinality, embedding_dim)"
            self.embedded_cat_dim = sum([t[1] for t in self.embedding_dims])
        else:
            self.embedded_cat_dim = 0
            

@dataclass
class ModelConfig:
    """Base Model configuration.

    Args:
        task (str): Specify whether the problem is regression or classification. `backbone` is a task which
                considers the model as a backbone to generate features. Mostly used internally for SSL and related
                tasks.. Choices are: [`regression`,`classification`,`backbone`].

        head (Optional[str]): The head to be used for the model. Should be one of the heads defined in
                `pytorch_tabular.models.common.heads`. Defaults to  LinearHead. Choices are:
                [`None`,`LinearHead`,`MixtureDensityHead`].

        head_config (Optional[Dict]): The config as a dict which defines the head. If left empty, will be
                initialized as default linear head.

        embedding_dims (Optional[List]): The dimensions of the embedding for each categorical column as a
                list of tuples (cardinality, embedding_dim). If left empty, will infer using the cardinality of
                the categorical column using the rule min(50, (x + 1) // 2)

        embedding_dropout (float): Dropout to be applied to the Categorical Embedding. Defaults to 0.0

        batch_norm_continuous_input (bool): If True, we will normalize the continuous layer by passing it
                through a BatchNorm layer.

        virtual_batch_size (Optional[int]): If not None, all BatchNorms will be converted to GhostBatchNorm's
                with the specified virtual batch size. Defaults to None

        loss (Optional[str]): The loss function to be applied. By Default, it is MSELoss for regression and
                CrossEntropyLoss for classification. Unless you are sure what you are doing, leave it at MSELoss
                or L1Loss for regression and CrossEntropyLoss for classification

    """

    task: str = field(
        metadata={
            "help": "Specify whether the problem is regression or classification."
            " `backbone` is a task which considers the model as a backbone to generate features."
            " Mostly used internally for SSL and related tasks.",
            "choices": ["regression", "classification", "backbone"],
        }
    )

    head: Optional[str] = field(
        default="LinearHead",
        metadata={
            "help": "The head to be used for the model. ",
            "choices": [None, "LinearHead", "MixtureDensityHead"],
        },
    )

    head_config: Optional[Dict] = field(
        default_factory=lambda: {"layers": ""},
        metadata={
            "help": "The config as a dict which defines the head."
            " If left empty, will be initialized as default linear head."
        },
    )
    embedding_dims: Optional[List] = field(
        default=None,
        metadata={
            "help": "The dimensions of the embedding for each categorical column as a list of tuples "
            "(cardinality, embedding_dim). If left empty, will infer using the cardinality of the "
            "categorical column using the rule min(50, (x + 1) // 2)"
        },
    )
    embedding_dropout: float = field(
        default=0.0,
        metadata={"help": "Dropout to be applied to the Categorical Embedding. Defaults to 0.0"},
    )
    batch_norm_continuous_input: bool = field(
        default=True,
        metadata={"help": "If True, we will normalize the continuous layer by passing it through a BatchNorm layer."},
    )
    loss: Optional[str] = field(
        default=None,
        metadata={
            "help": "The loss function to be applied. By Default it is MSELoss for regression "
            "and CrossEntropyLoss for classification. Unless you are sure what you are doing, "
            "leave it at MSELoss or L1Loss for regression and CrossEntropyLoss for classification"
        },
    )

    virtual_batch_size: Optional[int] = field(
        default=None,
        metadata={
            "help": "If not None, all BatchNorms will be converted to GhostBatchNorm's "
            " with this virtual batch size. Defaults to None"
        },
    )

    _module_src: str = field(default="models")
    _model_name: str = field(default="Model")
    _backbone_name: str = field(default="Backbone")
    _config_name: str = field(default="Config")

    def __post_init__(self):
        if self.task == "regression":
            self.loss = self.loss or "MSELoss"
        elif self.task =='binary':
            self.loss = self.loss or 'BCEWithLogitsLoss'
        elif self.task in ("multiclass","classification"):
            self.loss = self.loss or "CrossEntropyLoss"
        else:
            raise NotImplementedError(
                f"{self.task} is not a valid task. Should be one of "
                f"{self.__dataclass_fields__['task'].metadata['choices']}"
            )
        if self.task != "backbone":
            assert self.head in dir(heads.blocks), f"{self.head} is not a valid head"
            _head_callable = getattr(heads.blocks, self.head)
            ideal_head_config = _head_callable._config_template
            invalid_keys = set(self.head_config.keys()) - set(ideal_head_config.__dict__.keys())
            assert len(invalid_keys) == 0, f"`head_config` has some invalid keys: {invalid_keys}"

        # For Custom models, setting these values for compatibility
        if not hasattr(self, "_config_name"):
            self._config_name = type(self).__name__
        if not hasattr(self, "_model_name"):
            self._model_name = re.sub("[Cc]onfig", "Model", self._config_name)
        if not hasattr(self, "_backbone_name"):
            self._backbone_name = re.sub("[Cc]onfig", "Backbone", self._config_name)
            
    def merge_dataset_config(self,ds):
        inferred_config = get_inferred_config(ds)
        merged_config = safe_merge_config(OmegaConf.structured(self),
                        OmegaConf.structured(inferred_config))
        return merged_config
        




# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
"""AutomaticFeatureInteraction Config."""
from dataclasses import dataclass, field
from typing import Optional

from torchkeras.tabular.config import ModelConfig

@dataclass
class AutoIntConfig(ModelConfig):
    """AutomaticFeatureInteraction configuration.

    Args:
        attn_embed_dim (int): The number of hidden units in the Multi-Headed Attention layers. Defaults to
                32

        num_heads (int): The number of heads in the Multi-Headed Attention layer. Defaults to 2

        num_attn_blocks (int): The number of layers of stacked Multi-Headed Attention layers. Defaults to 3

        attn_dropouts (float): Dropout between layers of Multi-Headed Attention Layers. Defaults to 0.0

        has_residuals (bool): Flag to have a residual connect from embedded output to attention layer
                output. Defaults to True

        embedding_dim (int): The dimensions of the embedding for continuous and categorical columns.
                Defaults to 16

        embedding_initialization (Optional[str]): Initialization scheme for the embedding layers. Defaults
                to `kaiming`. Choices are: [`kaiming_uniform`,`kaiming_normal`].

        embedding_bias (bool): Flag to turn on Embedding Bias. Defaults to True

        share_embedding (bool): The flag turns on shared embeddings in the input embedding process. The key
                idea here is to have an embedding for the feature as a whole along with embeddings of each unique
                values of that column. For more details refer to Appendix A of the TabTransformer paper. Defaults
                to False

        share_embedding_strategy (Optional[str]): There are two strategies in adding shared embeddings. 1.
                `add` - A separate embedding for the feature is added to the embedding of the unique values of the
                feature. 2. `fraction` - A fraction of the input embedding is reserved for the shared embedding of
                the feature. Defaults to fraction. Choices are: [`add`,`fraction`].

        shared_embedding_fraction (float): Fraction of the input_embed_dim to be reserved by the shared
                embedding. Should be less than one. Defaults to 0.25

        deep_layers (bool): Flag to enable a deep MLP layer before the Multi-Headed Attention layer.
                Defaults to False

        layers (str): Hyphen-separated number of layers and units in the deep MLP. Defaults to 128-64-32

        activation (str): The activation type in the deep MLP. The default activation in PyTorch like ReLU,
                TanH, LeakyReLU, etc.
                https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity.
                Defaults to ReLU

        use_batch_norm (bool): Flag to include a BatchNorm layer after each Linear Layer+DropOut in the
                deep MLP. Defaults to False

        initialization (str): Initialization scheme for the linear layers in the deep MLP. Defaults to
                `kaiming`. Choices are: [`kaiming`,`xavier`,`random`].

        dropout (float): Probability of an element to be zeroed in the deep MLP. Defaults to 0.0

        attention_pooling (bool): If True, will combine the attention outputs of each block for final
                prediction. Defaults to False

        task (str): Specify whether the problem is regression or classification. `backbone` is a task which
                considers the model as a backbone to generate features. Mostly used internally for SSL and related
                tasks. Choices are: [`regression`,`classification`,`backbone`].

        head (Optional[str]): The head to be used for the model. Should be one of the heads defined in
                `pytorch_tabular.models.common.heads`. Defaults to  LinearHead. Choices are:
                [`None`,`LinearHead`,`MixtureDensityHead`].

        head_config (Optional[Dict]): The config as a dict which defines the head. If left empty, will be
                initialized as default linear head.

        embedding_dims (Optional[List]): The dimensions of the embedding for each categorical column as a
                list of tuples (cardinality, embedding_dim). If left empty, will infer using the cardinality of
                the categorical column using the rule min(50, (x + 1) // 2)

        embedding_dropout (float): Dropout to be applied to the Categorical Embedding. Defaults to 0.0

        batch_norm_continuous_input (bool): If True, we will normalize the continuous layer by passing it
                through a BatchNorm layer.

        loss (Optional[str]): The loss function to be applied. By Default, it is MSELoss for regression and
                CrossEntropyLoss for classification. Unless you are sure what you are doing, leave it at MSELoss
                or L1Loss for regression and CrossEntropyLoss for classification

    """

    attn_embed_dim: int = field(
        default=32,
        metadata={"help": "The number of hidden units in the Multi-Headed Attention layers. Defaults to 32"},
    )
    num_heads: int = field(
        default=2,
        metadata={"help": "The number of heads in the Multi-Headed Attention layer. Defaults to 2"},
    )
    num_attn_blocks: int = field(
        default=3,
        metadata={"help": "The number of layers of stacked Multi-Headed Attention layers. Defaults to 3"},
    )
    attn_dropouts: float = field(
        default=0.0,
        metadata={"help": "Dropout between layers of Multi-Headed Attention Layers. Defaults to 0.0"},
    )
    has_residuals: bool = field(
        default=True,
        metadata={
            "help": "Flag to have a residual connect from enbedded output to attention layer output. Defaults to True"
        },
    )
    embedding_dim: int = field(
        default=16,
        metadata={"help": "The dimensions of the embedding for continuous and categorical columns. Defaults to 16"},
    )
    embedding_initialization: Optional[str] = field(
        default="kaiming_uniform",
        metadata={
            "help": "Initialization scheme for the embedding layers. Defaults to `kaiming`",
            "choices": ["kaiming_uniform", "kaiming_normal"],
        },
    )
    embedding_bias: bool = field(
        default=True,
        metadata={"help": "Flag to turn on Embedding Bias. Defaults to True"},
    )
    share_embedding: bool = field(
        default=False,
        metadata={
            "help": "The flag turns on shared embeddings in the input embedding process."
            " The key idea here is to have an embedding for the feature as a whole along with embeddings"
            " of each unique values of that column."
            " For more details refer to Appendix A of the TabTransformer paper. Defaults to False"
        },
    )
    share_embedding_strategy: Optional[str] = field(
        default="fraction",
        metadata={
            "help": "There are two strategies in adding shared embeddings."
            " 1. `add` - A separate embedding for the feature is added to the embedding"
            " of the unique values of the feature."
            " 2. `fraction` - A fraction of the input embedding is reserved"
            " for the shared embedding of the feature. Defaults to fraction.",
            "choices": ["add", "fraction"],
        },
    )
    shared_embedding_fraction: float = field(
        default=0.25,
        metadata={
            "help": "Fraction of the input_embed_dim to be reserved by the shared embedding."
            " Should be less than one. Defaults to 0.25"
        },
    )
    deep_layers: bool = field(
        default=False,
        metadata={"help": "Flag to enable a deep MLP layer before the Multi-Headed Attention layer. Defaults to False"},
    )
    layers: str = field(
        default="128-64-32",
        metadata={"help": "Hyphen-separated number of layers and units in the deep MLP. Defaults to 128-64-32"},
    )
    activation: str = field(
        default="ReLU",
        metadata={
            "help": "The activation type in the deep MLP. The default activation in PyTorch"
            " like ReLU, TanH, LeakyReLU, etc."
            " https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity."
            " Defaults to ReLU"
        },
    )
    use_batch_norm: bool = field(
        default=False,
        metadata={
            "help": "Flag to include a BatchNorm layer after each Linear Layer+DropOut in the deep MLP."
            " Defaults to False"
        },
    )
    initialization: str = field(
        default="kaiming",
        metadata={
            "help": "Initialization scheme for the linear layers in the deep MLP. Defaults to `kaiming`",
            "choices": ["kaiming", "xavier", "random"],
        },
    )
    dropout: float = field(
        default=0.0,
        metadata={"help": "Probability of an element to be zeroed in the deep MLP. Defaults to 0.0"},
    )
    attention_pooling: bool = field(
        default=False,
        metadata={
            "help": "If True, will combine the attention outputs of each block for final prediction. Defaults to False"
        },
    )
    _module_src: str = field(default="models.autoint")
    _model_name: str = field(default="AutoIntModel")
    _backbone_name: str = field(default="AutoIntBackbone")
    _config_name: str = field(default="AutoIntConfig")




import torch
import torch.nn as nn
def _initialize_layers(activation, initialization, layers):
    if type(layers) is nn.Sequential:
        for layer in layers:
            if hasattr(layer, "weight"):
                _initialize_layers(activation, initialization, layer)
    else:
        if activation == "ReLU":
            nonlinearity = "relu"
        elif activation == "LeakyReLU":
            nonlinearity = "leaky_relu"
        else:
            if initialization == "kaiming":
                nonlinearity = "leaky_relu"
            else:
                nonlinearity = "relu"

        if initialization == "kaiming":
            nn.init.kaiming_normal_(layers.weight, nonlinearity=nonlinearity)
        elif initialization == "xavier":
            nn.init.xavier_normal_(
                layers.weight,
                gain=(nn.init.calculate_gain(nonlinearity) if activation in ["ReLU", "LeakyReLU"] else 1),
            )
        elif initialization == "random":
            nn.init.normal_(layers.weight)


def _linear_dropout_bn(activation, initialization, use_batch_norm, in_units, out_units, dropout):
    if isinstance(activation, str):
        _activation = getattr(nn, activation)
    else:
        _activation = activation
    layers = []
    if use_batch_norm:
        from torchkeras.tabular.models.common.layers.batch_norm import BatchNorm1d

        layers.append(BatchNorm1d(num_features=in_units))
    linear = nn.Linear(in_units, out_units)
    _initialize_layers(activation, initialization, linear)
    layers.extend([linear, _activation()])
    if dropout != 0:
        layers.append(nn.Dropout(dropout))
    return layers


def reset_all_weights(model: nn.Module) -> None:
    """Resets all parameters in a network.

    Args:
        model: The model to reset the parameters of.

    refs:
        - https://discuss.pytorch.org/t/how-to-re-set-alll-parameters-in-a-network/20819/6
        - https://stackoverflow.com/questions/63627997/reset-parameters-of-a-neural-network-in-pytorch
        - https://pytorch.org/docs/stable/generated/torch.nn.Module.html

    """

    @torch.no_grad()
    def weight_reset(m: nn.Module):
        # - check if the current module has reset_parameters & if it's callabed called it on m
        reset_parameters = getattr(m, "reset_parameters", None)
        if callable(reset_parameters):
            m.reset_parameters()

    # Applies fn recursively to every submodule see: https://pytorch.org/docs/stable/generated/torch.nn.Module.html
    model.apply(fn=weight_reset)


def to_one_hot(y, depth=None):
    r"""Takes integer with n dims and converts it to 1-hot representation with n + 1 dims.

    The n+1'st dimension will have zeros everywhere but at y'th index, where it will be equal to 1.
    Args:
        y: input integer (IntTensor, LongTensor or Variable) of any shape
        depth (int):  the size of the one hot dimension

    """
    y_flat = y.to(torch.int64).view(-1, 1)
    depth = depth or int(torch.max(y_flat)) + 1
    y_one_hot = torch.zeros(y_flat.size()[0], depth, device=y.device).scatter_(1, y_flat, 1)
    y_one_hot = y_one_hot.view(*(tuple(y.shape) + (-1,)))
    return y_one_hot


def _make_ix_like(input, dim=0):
    d = input.size(dim)
    rho = torch.arange(1, d + 1, device=input.device, dtype=input.dtype)
    view = [1] * input.dim()
    view[0] = -1
    return rho.view(view).transpose(0, dim)


def _initialize_kaiming(x, initialization, d_sqrt_inv):
    if initialization == "kaiming_uniform":
        nn.init.uniform_(x, a=-d_sqrt_inv, b=d_sqrt_inv)
    elif initialization == "kaiming_normal":
        nn.init.normal_(x, std=d_sqrt_inv)
    elif initialization is None:
        pass
    else:
        raise NotImplementedError("initialization should be either of `kaiming_normal`, `kaiming_uniform`," " `None`")


class OutOfMemoryHandler:
    """Context manager to handle out of memory errors.

    Args:
        handle_oom: Whether to handle the error or not. If set to False,
            the exception will be propagated.

    """

    def __init__(self, handle_oom: bool = True):
        self.handle_oom = handle_oom
        self.oom_triggered = False
        self.oom_msg = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        is_oom_runtime_error = exc_type is RuntimeError and "out of memory" in str(exc_value)
        try:
            is_cuda_oom_error = exc_type is torch.cuda.OutOfMemoryError
        except AttributeError:
            # before torch 1.13.0, torch.cuda.OutOfMemoryError did not exist
            is_cuda_oom_error = False
        if (is_oom_runtime_error or is_cuda_oom_error) and self.handle_oom:
            self.oom_triggered = True
            self.oom_msg = exc_value.args[0]
            torch.cuda.empty_cache()
            return True  # Suppress the exception
        return False  # Propagate any other exceptions


class OOMException(Exception):
    pass


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)



# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
"""AutomaticFeatureInteraction Config."""
from dataclasses import dataclass, field

from torchkeras.tabular.config import ModelConfig

@dataclass
class GANDALFConfig(ModelConfig):
    """Gated Adaptive Network for Deep Automated Learning of Features (GANDALF) Config.

    Args:
        gflu_stages (int): Number of layers in the feature abstraction layer. Defaults to 6

        gflu_dropout (float): Dropout rate for the feature abstraction layer. Defaults to 0.0

        gflu_feature_init_sparsity (float): Only valid for t-softmax. The percentage of features
                to be selected in each GFLU stage. This is just initialized and during learning
                it may change. Defaults to 0.3

        learnable_sparsity (bool): Only valid for t-softmax. If True, the sparsity parameters
                will be learned. If False, the sparsity parameters will be fixed to the initial
                values specified in `gflu_feature_init_sparsity` and `tree_feature_init_sparsity`.
                Defaults to True

        task (str): Specify whether the problem is regression or classification. `backbone` is a task which
                considers the model as a backbone to generate features. Mostly used internally for SSL and related
                tasks. Choices are: [`regression`,`classification`,`backbone`].

        head (Optional[str]): The head to be used for the model. Should be one of the heads defined in
                `pytorch_tabular.models.common.heads`. Defaults to  LinearHead. Choices are:
                [`None`,`LinearHead`,`MixtureDensityHead`].

        head_config (Optional[Dict]): The config as a dict which defines the head. If left empty, will be
                initialized as default linear head.

        embedding_dims (Optional[List]): The dimensions of the embedding for each categorical column as a
                list of tuples (cardinality, embedding_dim). If left empty, will infer using the cardinality of
                the categorical column using the rule min(50, (x + 1) // 2)

        embedding_dropout (float): Dropout to be applied to the Categorical Embedding. Defaults to 0.0

        batch_norm_continuous_input (bool): If True, we will normalize the continuous layer by passing it
                through a BatchNorm layer.

        loss (Optional[str]): The loss function to be applied. By Default, it is MSELoss for regression and
                CrossEntropyLoss for classification. Unless you are sure what you are doing, leave it at MSELoss
                or L1Loss for regression and CrossEntropyLoss for classification

    """

    gflu_stages: int = field(
        default=6,
        metadata={"help": "Number of layers in the feature abstraction layer. Defaults to 6"},
    )

    gflu_dropout: float = field(
        default=0.0, metadata={"help": "Dropout rate for the feature abstraction layer. Defaults to 0.0"}
    )

    gflu_feature_init_sparsity: float = field(
        default=0.3,
        metadata={
            "help": "Only valid for t-softmax. The perecentge of features to be selected in "
            "each GFLU stage. This is just initialized and during learning it may change"
        },
    )
    learnable_sparsity: bool = field(
        default=True,
        metadata={
            "help": "Only valid for t-softmax. If True, the sparsity parameters will be learned."
            "If False, the sparsity parameters will be fixed to the initial values specified in "
            "`gflu_feature_init_sparsity` and `tree_feature_init_sparsity`"
        },
    )
    _module_src: str = field(default="models.gandalf")
    _model_name: str = field(default="GANDALFModel")
    _backbone_name: str = field(default="GANDALFBackbone")
    _config_name: str = field(default="GANDALFConfig")

    def __post_init__(self):
        assert self.gflu_stages > 0, "gflu_stages should be greater than 0"
        return super().__post_init__()




from .config import GatedAdditiveTreeEnsembleConfig
from .gate_model import GatedAdditiveTreeEnsembleModel, GatedAdditiveTreesBackbone

__all__ = [
    "GatedAdditiveTreesBackbone",
    "GatedAdditiveTreeEnsembleModel",
    "GatedAdditiveTreeEnsembleConfig",
]



from . import (
    autoint,
    category_embedding,
    danet,
    ft_transformer,
    gandalf,
    gate,
    node,
    tab_transformer,
    tabnet,
    fm,
    deepfm,
    deepcross,
    pemlp,
    deepknn
)

from .autoint import AutoIntConfig, AutoIntModel
from .base_model import BaseModel
from .category_embedding import CategoryEmbeddingModel, CategoryEmbeddingModelConfig
from .danet import DANetConfig, DANetModel
from .ft_transformer import FTTransformerConfig, FTTransformerModel
from .gandalf import GANDALFBackbone, GANDALFConfig, GANDALFModel
from .gate import GatedAdditiveTreeEnsembleConfig, GatedAdditiveTreeEnsembleModel
from .node import NODEConfig, NODEModel
from .tab_transformer import TabTransformerConfig, TabTransformerModel
from .tabnet import TabNetModel, TabNetModelConfig
from .fm import FMModel, FMConfig
from .deepfm import DeepFMModel, DeepFMConfig
from .deepcross import DeepCrossModel, DeepCrossConfig
from .pemlp import PeMLPModel, PeMLPConfig
from .deepknn import DeepKNNModel, DeepKNNConfig

__all__ = [
    "BaseModel",
    
    "CategoryEmbeddingModel",
    "CategoryEmbeddingModelConfig",
    "category_embedding",
    
    "NODEModel",
    "NODEConfig",
    "node",
    
    "TabNetModel",
    "TabNetModelConfig",
    "tabnet",
    
    "AutoIntConfig",
    "AutoIntModel",
    "autoint",
    
    "TabTransformerConfig",
    "TabTransformerModel",
    "tab_transformer",
    
    "FTTransformerConfig",
    "FTTransformerModel",
    "ft_transformer",
    
    "GatedAdditiveTreeEnsembleConfig",
    "GatedAdditiveTreeEnsembleModel",
    "gate",
    
    "GANDALFConfig",
    "GANDALFModel",
    "GANDALFBackbone",
    "gandalf",
    
    "DANetConfig",
    "DANetModel",
    "danet",

    "FMConfig",
    "FMModel",    
    "fm",

    "DeepFMConfig",
    "DeepFMModel",
    "deepfm",

    "DeepCrossConfig",
    "DeepCrossModel",
    "deepcross",

    "PeMLPConfig",
    "PeMLPModel",
    "pemlp",

    "DeepKNNConfig",
    "DeepKNNModel",
    "deepknn",

    
]



from torchkeras import VLog

class VLogCallback:
    def __init__(self, num_boost_round, 
                 monitor_metric='val_loss',
                 monitor_mode='min'):
        self.order = 20
        self.num_boost_round = num_boost_round
        self.vlog = VLog(epochs = num_boost_round, monitor_metric = monitor_metric, 
                         monitor_mode = monitor_mode)

    def __call__(self, env) -> None:
        metrics = {}
        for item in env.evaluation_result_list:
            if len(item) == 4:
                data_name, eval_name, result = item[:3]
                metrics[data_name+'_'+eval_name] = result
            else:
                data_name, eval_name = item[1].split()
                res_mean = item[2]
                res_stdv = item[4]
                metrics[data_name+'_'+eval_name] = res_mean
        self.vlog.log_epoch(metrics)
        
if __name__=='__main__':
    import datetime
    import numpy as np
    import pandas as pd
    import lightgbm as lgb
    from sklearn import datasets
    from sklearn.model_selection import train_test_split
    from sklearn.metrics import accuracy_score
    from torchkeras.tools.lightgbm import VLogCallback 

    def printlog(info):
        nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        print("\n"+"=========="*8 + "%s"%nowtime)
        print(info+'...\n\n')


    #================================================================================
    printlog("step1: reading data...")


    breast = datasets.load_breast_cancer()
    df = pd.DataFrame(breast.data,columns = [x.replace(' ','_') for x in breast.feature_names])
    df['label'] = breast.target
    df['mean_radius'] = df['mean_radius'].apply(lambda x:int(x))
    df['mean_texture'] = df['mean_texture'].apply(lambda x:int(x))
    dftrain,dftest = train_test_split(df)

    categorical_features = ['mean_radius','mean_texture']
    lgb_train = lgb.Dataset(dftrain.drop(['label'],axis = 1),label=dftrain['label'],
                            categorical_feature = categorical_features)

    lgb_valid = lgb.Dataset(dftest.drop(['label'],axis = 1),label=dftest['label'],
                            categorical_feature = categorical_features,
                            reference=lgb_train)


    #================================================================================
    printlog("step2: setting parameters...")

    boost_round = 50                   
    early_stop_rounds = 10

    params = {
        'boosting_type': 'gbdt',
        'objective':'binary',
        'metric': ['auc'], #'l2'
        'num_leaves': 15,   
        'learning_rate': 0.05,
        'feature_fraction': 0.9,
        'bagging_fraction': 0.8,
        'bagging_freq': 5,
        'verbose': 0,
        'early_stopping_round':5
    }


    #================================================================================
    printlog("step3: training model...")

    result = {}

    vlog_cb = VLogCallback(boost_round, monitor_metric = 'val_auc', monitor_mode = 'max')
    vlog_cb.vlog.log_start()

    gbm = lgb.train(params,
                    lgb_train,
                    num_boost_round= boost_round,
                    valid_sets=(lgb_valid, lgb_train),
                    valid_names=('val','train'),
                    callbacks = [lgb.record_evaluation(result),
                                 vlog_cb]
                   )

    vlog_cb.vlog.log_end()

 
    #================================================================================
    printlog("step4: evaluating model ...")

    y_pred_train = gbm.predict(dftrain.drop('label',axis = 1), num_iteration=gbm.best_iteration)
    y_pred_test = gbm.predict(dftest.drop('label',axis = 1), num_iteration=gbm.best_iteration)

    print('train accuracy: {:.5} '.format(accuracy_score(dftrain['label'],y_pred_train>0.5)))
    print('valid accuracy: {:.5} \n'.format(accuracy_score(dftest['label'],y_pred_test>0.5)))

    lgb.plot_metric(result,metric='auc')
    lgb.plot_importance(gbm,importance_type = "gain")



    #================================================================================
    printlog("step5: saving model ...")

    model_dir = "gbm.model"
    print("model_dir: %s"%model_dir)
    gbm.save_model("gbm.model")
    printlog("task end...")



from torchkeras import VLog
from pytorch_lightning.callbacks import Callback
class VLogCallback(Callback):
    def __init__(self, monitor_metric='val_loss', monitor_mode='min'):
        super().__init__()
        self.monitor_metric = monitor_metric
        self.monitor_mode = monitor_mode 
    
    def on_fit_start(self, trainer, pl_module):
        self.vlog = VLog(trainer.max_epochs,
                         monitor_metric=self.monitor_metric, 
                         monitor_mode=self.monitor_mode
                        ) 
        self.vlog.log_start() 
        
    def on_validation_epoch_end(self, trainer, pl_module):
        if trainer.current_epoch>0:
            row_data = {k:v for k,v in trainer.logged_metrics.items() 
                        if k.split('_')[-1]==self.monitor_metric.split('_')[-1]
                       }
            self.vlog.log_epoch(row_data) 
            
    def on_fit_end(self,trainer,pl_module):
        self.vlog.log_end()


import torch 
from torch import nn
from torch import nn,Tensor 
import torch.nn.functional as F 

from omegaconf import DictConfig
from typing import Dict,Any
from ..base_model import BaseModel

class NumEmbedding(nn.Module):
    """
    input_shape: [batch_size,features_num(n), d_in], # d_in=1 usually 
    output_shape: [batch_size,features_num(n), d_out]
    """
    
    def __init__(self, n: int, d_in: int, d_out: int, bias: bool = False) -> None:
        super().__init__()
        self.weight = nn.Parameter(Tensor(n, d_in, d_out))
        self.bias = nn.Parameter(Tensor(n, d_out)) if bias else None
        with torch.no_grad():
            for i in range(n):
                layer = nn.Linear(d_in, d_out)
                self.weight[i] = layer.weight.T
                if self.bias is not None:
                    self.bias[i] = layer.bias

    def forward(self, x_num):
        # x_num: batch_size, features_num, d_in
        assert x_num.ndim == 3
        x = x_num[..., None] * self.weight[None]
        x = x.sum(-2)
        #x = torch.einsum("bfi,fij->bfj",x_num,self.weight)
        if self.bias is not None:
            x = x + self.bias[None]
        return x
    
class CatEmbedding(nn.Module):
    """
    input_shape: [batch_size,features_num], 
    output_shape: [batch_size,features_num, d_embed]
    """
    def __init__(self, categories, d_embed):
        super().__init__()
        self.embedding = nn.Embedding(sum(categories), d_embed)
        self.offsets = nn.Parameter(
                torch.tensor([0] + categories[:-1]).cumsum(0),requires_grad=False)
        
        torch.nn.init.xavier_uniform_(self.embedding.weight.data)

    def forward(self, x_cat):
        """
        :param x_cat: Long tensor of size ``(batch_size, features_num)``
        """
        x = x_cat + self.offsets[None]
        return self.embedding(x) 
    
class CatLinear(nn.Module):
    """
    equals F.onehot + nn.Linear
    input_shape: [batch_size,features_num], 
    output_shape: [batch_size,d_out]
    """
    def __init__(self, categories, d_out=1):
        super().__init__()
        self.fc = nn.Embedding(sum(categories), d_out)
        self.bias = nn.Parameter(torch.zeros((d_out,)))
        self.offsets = nn.Parameter(
                torch.tensor([0] + categories[:-1]).cumsum(0),requires_grad=False)

    def forward(self, x_cat):
        """
        :param x: Long tensor of size ``(batch_size, features_num)``
        """
        x = x_cat + self.offsets[None]
        return torch.sum(self.fc(x), dim=1) + self.bias 
    
    
class FMLayer(nn.Module):

    def __init__(self, reduce_sum=True):
        super().__init__()
        self.reduce_sum = reduce_sum

    def forward(self, x): #注意：这里的x是公式中的 <v_i> * xi
        """
        :param x: Float tensor of size ``(batch_size, num_features, k)``
        """
        square_of_sum = torch.sum(x, dim=1) ** 2
        sum_of_square = torch.sum(x ** 2, dim=1)
        ix = square_of_sum - sum_of_square
        if self.reduce_sum:
            ix = torch.sum(ix, dim=1, keepdim=True)
        return 0.5 * ix

class MultiLayerPerceptron(nn.Module):

    def __init__(self, d_in, d_layers, dropout, 
                 d_out = 1):
        super().__init__()
        layers = []
        for d in d_layers:
            layers.append(nn.Linear(d_in, d))
            layers.append(nn.BatchNorm1d(d))
            layers.append(nn.ReLU())
            layers.append(nn.Dropout(p=dropout))
            d_in = d
        layers.append(nn.Linear(d_layers[-1], d_out))
        self.mlp = nn.Sequential(*layers)

    def forward(self, x):
        """
        :param x: Float tensor of size ``(batch_size, d_in)``
        """
        return self.mlp(x)
    


class DeepFMBackbone(nn.Module):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__()
        self.hparams = config
        
        d_numerical = self.hparams.continuous_dim
        d_embed = self.hparams.input_embed_dim
        categories = self.hparams.categorical_cardinality
        n_classes = self.hparams.output_dim 
        
        deep_layers = [int(x) for x in self.hparams.deep_layers.split('-')]
        deep_dropout = self.hparams.deep_dropout

        
        if d_numerical is None:
            d_numerical = 0
            
        if categories is None:
            categories = []
            
        self.categories = categories
        self.n_classes = n_classes
        
        self.num_linear = nn.Linear(d_numerical,n_classes) if d_numerical else None
        self.cat_linear = CatLinear(categories,n_classes) if categories else None
        
        self.num_embedding = NumEmbedding(d_numerical,1,d_embed) if d_numerical else None
        self.cat_embedding = CatEmbedding(categories, d_embed) if categories else None
        

        self.fm = FMLayer(reduce_sum=False)
        self.fm_linear = nn.Linear(d_embed,n_classes)
            
        self.deep_in = d_numerical*d_embed+len(categories)*d_embed
        
        self.deep = MultiLayerPerceptron(
            d_in= self.deep_in,
            d_layers = deep_layers,
            dropout = deep_dropout,
            d_out = n_classes
        )
        

    def forward(self, x: Dict):
        x_cat,x_num = x["categorical"], x["continuous"]
        
        #linear部分
        x = 0.0
        if self.num_linear:
            x = x + self.num_linear(x_num) 
        if self.cat_linear:
            x = x + self.cat_linear(x_cat)
        
        #fm部分
        x_embedding = []
        if self.num_embedding:
            x_embedding.append(self.num_embedding(x_num[...,None]))
        if self.cat_embedding:
            x_embedding.append(self.cat_embedding(x_cat))
        x_embedding = torch.cat(x_embedding,dim=1)
        

        x = x + self.fm_linear(self.fm(x_embedding)) 
            
        #deep部分
        x = x + self.deep(x_embedding.view(-1,self.deep_in))
        
        if self.n_classes==1:
            x = x.squeeze(-1)
        return x
        

class DeepFMModel(BaseModel):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__(config, **kwargs)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head
        
    def _build_network(self):
        self._embedding_layer = nn.Identity()
        self._backbone = DeepFMBackbone(self.hparams)
        setattr(self.backbone, "output_dim", self.hparams.output_dim)
        self._head = nn.Identity()
        
    def forward(self, x: Dict) -> Dict[str, Any]:
        x = self.embed_input(x)
        x = self.compute_backbone(x)
        return self.compute_head(x)

    def extract_embedding(self):
        raise ValueError("Extracting Embeddings is not supported by DeepFMModel.")
        





from .data_utils import (
    get_balanced_sampler,
    get_class_weighted_cross_entropy,
    get_gaussian_centers,
    load_covertype_dataset,
    make_mixed_dataset,
    print_metrics,
)

from .nn_utils import (
    OOMException,
    OutOfMemoryHandler,
    _initialize_kaiming,
    _initialize_layers,
    _linear_dropout_bn,
    _make_ix_like,
    count_parameters,
    reset_all_weights,
    to_one_hot,
)
from .python_utils import (
    available_models,
    check_numpy,
    generate_doc_dataclass,
    getattr_nested,
    ifnone,
    int_to_human_readable,
)

__all__ = [
    "getattr_nested",
    "generate_doc_dataclass",
    "ifnone",
    "_initialize_layers",
    "_linear_dropout_bn",
    "reset_all_weights",
    "get_class_weighted_cross_entropy",
    "get_balanced_sampler",
    "get_gaussian_centers",
    "_make_ix_like",
    "to_one_hot",
    "_initialize_kaiming",
    "check_numpy",
    "OutOfMemoryHandler",
    "OOMException",
    "make_mixed_dataset",
    "print_metrics",
    "load_covertype_dataset",
    "count_parameters",
    "int_to_human_readable",
    "available_models"
]



import os 
import sys
import datetime 
from copy import deepcopy
import numpy as np 
import pandas as pd 
from argparse import Namespace 
from torchkeras.utils import is_jupyter

class TensorBoardCallback:
    def __init__(self, save_dir="runs", model_name="model",
                 log_weight=False, log_weight_freq=5):
        """
        TensorBoard callback for logging training progress

        Args:
        -  save_dir (str): Directory to save TensorBoard logs
        -  model_name (str): Name of the model
        -  log_weight (bool): Whether to log model weights
        -  log_weight_freq (int): Frequency of logging model weights during training
        """
        from torch.utils.tensorboard import SummaryWriter
        self.__dict__.update(locals())
        nowtime = datetime.datetime.now().strftime('%Y%m%d_%H%M%S')
        self.log_path = os.path.join(save_dir, model_name, nowtime)
        self.writer = SummaryWriter(self.log_path)

    def on_fit_start(self, model: 'KerasModel'):
        """
        Callback function called at the beginning of model fitting

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        # Log model weights
        if self.log_weight:
            net = model.accelerator.unwrap_model(model.net)
            for name, param in net.named_parameters():
                self.writer.add_histogram(name, param.clone().cpu().data.numpy(), 0)
            self.writer.flush()

    def on_train_epoch_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of each training epoch

        Args:
        - model (KerasModel): The KerasModel being trained.
        """
        epoch = max(model.history['epoch'])

        # Log model weights
        net = model.accelerator.unwrap_model(model.net)
        if self.log_weight and epoch % self.log_weight_freq == 0:
            for name, param in net.named_parameters():
                self.writer.add_histogram(name, param.clone().cpu().data.numpy(), epoch)
            self.writer.flush()

    def on_validation_epoch_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of each validation epoch

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        dfhistory = pd.DataFrame(model.history)
        n = len(dfhistory)
        epoch = max(model.history['epoch'])

        # Log metrics
        dic = deepcopy(dfhistory.iloc[n - 1])
        dic.pop("epoch")

        metrics_group = {}
        for key, value in dic.items():
            g = key.replace("train_", '').replace("val_", '')
            metrics_group[g] = dict(metrics_group.get(g, {}), **{key: value})
        for group, metrics in metrics_group.items():
            self.writer.add_scalars(group, metrics, epoch)
        self.writer.flush()

    def on_fit_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of model fitting

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        # Log model weights
        epoch = max(model.history['epoch'])
        if self.log_weight:
            net = model.accelerator.unwrap_model(model.net)
            for name, param in net.named_parameters():
                self.writer.add_histogram(name, param.clone().cpu().data.numpy(), epoch)
            self.writer.flush()
        self.writer.close()

        # Save history
        dfhistory = pd.DataFrame(model.history)
        dfhistory.to_csv(os.path.join(self.log_path, 'dfhistory.csv'), index=None)


class WandbCallback:
    def __init__(self, project=None, config=None, name=None, save_ckpt=True, save_code=True):
        """
        WandbCallback for logging training progress using Weights & Biases

        Args:
        - project (str): Name of the project in W&B
        - config (dict or Namespace): Configuration parameters
        - name (str): Name of the run.
        - save_ckpt (bool): Whether to save model checkpoints
        - save_code (bool): Whether to save code artifacts
        """
        self.__dict__.update(locals())
        if isinstance(config, Namespace):
            self.config = config.__dict__
        if name is None:
            self.name = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        import wandb
        self.wb = wandb

    def on_fit_start(self, model: 'KerasModel'):
        """
        Callback function called at the beginning of model fitting

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        if self.wb.run is None:
            self.wb.init(project=self.project, config=self.config, name=self.name, save_code=self.save_code)
        model.run_id = self.wb.run.id

    def on_train_epoch_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of each training epoch

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        pass

    def on_validation_epoch_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of each validation epoch

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        dfhistory = pd.DataFrame(model.history)
        n = len(dfhistory)
        if n == 1:
            for m in dfhistory.columns:
                self.wb.define_metric(name=m, step_metric='epoch', hidden=False if m != 'epoch' else True)
            self.wb.define_metric(name='best_' + model.monitor, step_metric='epoch')

        dic = dict(dfhistory.iloc[n - 1])
        monitor_arr = dfhistory[model.monitor]
        best_monitor_score = monitor_arr.max() if model.mode == 'max' else monitor_arr.min()
        dic.update({'best_' + model.monitor: best_monitor_score})
        self.wb.run.summary["best_score"] = best_monitor_score
        self.wb.log(dic)

    def on_fit_end(self, model: 'KerasModel'):
        """
        Callback function called at the end of model fitting

        Args:
        - model (KerasModel): The KerasModel being trained
        """
        # Save dfhistory
        dfhistory = pd.DataFrame(model.history)
        dfhistory.to_csv(os.path.join(self.wb.run.dir, 'dfhistory.csv'), index=None)

        # Save ckpt
        if self.save_ckpt:
            arti_model = self.wb.Artifact('checkpoint', type='model')
            if os.path.isdir(model.ckpt_path):
                arti_model.add_dir(model.ckpt_path)
            else:
                arti_model.add_file(model.ckpt_path)
            self.wb.log_artifact(arti_model)

        run_dir = self.wb.run.dir
        self.wb.finish()

        # Local save
        try:
            import shutil
            copy_fn = shutil.copytree if os.path.isdir(model.ckpt_path) else shutil.copy
            copy_fn(model.ckpt_path, os.path.join(run_dir, os.path.basename(model.ckpt_path)))
        except Exception as err:
            print(err)


class VisProgress:
    def __init__(self):
        pass

    def on_fit_start(self, model: 'KerasModel'):
        """Callback at the beginning of the training

        Args:
            model (KerasModel): The KerasModel instance
        """
        from .pbar import ProgressBar
        self.progress = ProgressBar(range(model.epochs))
        model.EpochRunner.progress = self.progress

    def on_train_epoch_end(self, model: 'KerasModel'):
        """Callback at the end of each training epoch.

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass

    def on_validation_epoch_end(self, model: "KerasModel"):
        """Callback at the end of each validation epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        dfhistory = pd.DataFrame(model.history)
        self.progress.update(dfhistory['epoch'].iloc[-1])

    def on_fit_end(self, model: "KerasModel"):
        """Callback at the end of the entire training process

        Args:
            model (KerasModel): The KerasModel instance
        """
        dfhistory = pd.DataFrame(model.history)
        if dfhistory['epoch'].max() < model.epochs:
            self.progress.on_interrupt(msg='')
        self.progress.display = False


class VisMetric:
    def __init__(self, figsize=(6, 4), save_path='history.png'):
        """Visualization callback for monitoring metrics

        Args:
            figsize (tuple, optional): Figure size. Defaults to (6, 4)
            save_path (str, optional): Path to save the history plot. Defaults to 'history.png'
        """
        self.figsize = figsize
        self.save_path = save_path
        self.in_jupyter = is_jupyter()

    def on_fit_start(self, model: 'KerasModel'):
        """Callback at the beginning of the training

        Args:
            model (KerasModel): The KerasModel instance.
        """
        if not self.in_jupyter:
            print('\nView dynamic loss/metric plot: \n' + os.path.abspath(self.save_path))
        self.metric = model.monitor.replace('val_', '')
        dfhistory = pd.DataFrame(model.history)
        x_bounds = [0, min(10, model.epochs)]
        title = f'best {model.monitor} = ?'
        self.update_graph(model, title=title, x_bounds=x_bounds)

    def on_train_epoch_end(self, model: 'KerasModel'):
        """Callback at the end of each training epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass

    def on_validation_epoch_end(self, model: "KerasModel"):
        """Callback at the end of each validation epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        dfhistory = pd.DataFrame(model.history)
        n = len(dfhistory)
        x_bounds = [dfhistory['epoch'].min(), min(10 + (n // 10) * 10, model.epochs)]
        title = self.get_title(model)
        self.update_graph(model, title=title, x_bounds=x_bounds)

    def on_fit_end(self, model: "KerasModel"):
        """Callback at the end of the entire training process

        Args:
            model (KerasModel): The KerasModel instance
        """
        dfhistory = pd.DataFrame(model.history)
        title = self.get_title(model)
        self.update_graph(model, title=title)

    def get_best_score(self, model: 'KerasModel'):
        """Get the best score and epoch.

        Args:
            model (KerasModel): The KerasModel instance

        Returns:
            tuple: Best epoch and best score
        """
        dfhistory = pd.DataFrame(model.history)
        arr_scores = dfhistory[model.monitor]
        best_score = np.max(arr_scores) if model.mode == "max" else np.min(arr_scores)
        best_epoch = dfhistory.loc[arr_scores == best_score, 'epoch'].tolist()[0]
        return (best_epoch, best_score)

    def get_title(self, model: 'KerasModel'):
        """Get the title for the plot

        Args:
            model (KerasModel): The KerasModel instance

        Returns:
            str: The title.
        """
        best_epoch, best_score = self.get_best_score(model)
        title = f'best {model.monitor}={best_score:.4f} (@epoch {best_epoch})'
        return title

    def update_graph(self, model: 'KerasModel', title=None, x_bounds=None, y_bounds=None):
        """Update the metric plot.

        Args:
            model (KerasModel): The KerasModel instance
            title (str, optional): Plot title. Defaults to None
            x_bounds (list, optional): x-axis bounds. Defaults to None
            y_bounds (list, optional): y-axis bounds. Defaults to None
        """
        import matplotlib.pyplot as plt
        self.plt = plt
        if not hasattr(self, 'graph_fig'):
            self.graph_fig, self.graph_ax = plt.subplots(1, figsize=self.figsize)
            if self.in_jupyter:
                self.graph_out = display(self.graph_ax.figure, display_id=True)
        self.graph_ax.clear()

        dfhistory = pd.DataFrame(model.history)
        epochs = dfhistory['epoch'] if 'epoch' in dfhistory.columns else []

        m1 = "train_" + self.metric
        if m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            self.graph_ax.plot(epochs, train_metrics, 'bo--', label=m1, clip_on=False)

        m2 = 'val_' + self.metric
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.graph_ax.plot(epochs, val_metrics, 'co-', label=m2, clip_on=False)

        if self.metric in dfhistory.columns:
            metric_values = dfhistory[self.metric]
            self.graph_ax.plot(epochs, metric_values, 'co-', label=self.metric, clip_on=False)

        self.graph_ax.set_xlabel("epoch")
        self.graph_ax.set_ylabel(self.metric)
        if title:
            self.graph_ax.set_title(title)
            if not self.in_jupyter and hasattr(model.EpochRunner, 'progress'):
                model.EpochRunner.progress.comment_tail = title
        if m1 in dfhistory.columns or m2 in dfhistory.columns or self.metric in dfhistory.columns:
            self.graph_ax.legend(loc='best')

        if len(epochs) > 0:
            best_epoch, best_score = self.get_best_score(model)
            self.graph_ax.plot(best_epoch, best_score, 'r*', markersize=15, clip_on=False)

        if x_bounds is not None: self.graph_ax.set_xlim(*x_bounds)
        if y_bounds is not None: self.graph_ax.set_ylim(*y_bounds)
        if self.in_jupyter:
            self.graph_out.update(self.graph_ax.figure)
        self.graph_fig.savefig(self.save_path)
        self.plt.close()

class VisDisplay:
    def __init__(self, display_fn, model=None, init_display=True, dis_period=1):
        """Visualization callback for displaying custom information during training

        Args:
            display_fn (callable): Function to display information. Should accept a KerasModel instance
            model (KerasModel, optional): The KerasModel instance. Defaults to None
            init_display (bool, optional): Whether to display information initially. Defaults to True
            dis_period (int, optional): Display period (in epochs). Defaults to 1
        """
        from ipywidgets import Output
        self.display_fn = display_fn
        self.init_display = init_display
        self.dis_period = dis_period
        self.out = Output()

        if self.init_display:
            display(self.out)
            with self.out:
                self.display_fn(model)

    def on_fit_start(self, model: 'KerasModel'):
        """Callback at the beginning of the training

        Args:
            model (KerasModel): The KerasModel instance
        """
        if not self.init_display:
            display(self.out)

    def on_train_epoch_end(self, model: 'KerasModel'):
        """Callback at the end of each training epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass

    def on_validation_epoch_end(self, model: "KerasModel"):
        """Callback at the end of each validation epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        if len(model.history['epoch']) % self.dis_period == 0:
            self.out.clear_output()
            with self.out:
                self.display_fn(model)

    def on_fit_end(self, model: "KerasModel"):
        """Callback at the end of the entire training process

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass


class EpochCheckpoint:
    def __init__(self, ckpt_dir="weights", save_freq=1, max_ckpt=10):
        """Callback for saving model checkpoints during training

        Args:
            ckpt_dir (str, optional): Directory to save checkpoints. Defaults to "weights"
            save_freq (int, optional): Save frequency (in epochs). Defaults to 1
            max_ckpt (int, optional): Maximum number of checkpoints to keep. Defaults to 10
        """
        self.__dict__.update(locals())
        self.ckpt_idx = 0

    def on_fit_start(self, model: 'KerasModel'):
        """Callback at the beginning of the training

        Args:
            model (KerasModel): The KerasModel instance
        """
        if not os.path.exists(self.ckpt_dir):
            os.mkdir(self.ckpt_dir)
        self.ckpt_list = ['' for i in range(self.max_ckpt)]

    def on_train_epoch_end(self, model: 'KerasModel'):
        """Callback at the end of each training epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass

    def on_validation_epoch_end(self, model: "KerasModel"):
        """Callback at the end of each validation epoch

        Args:
            model (KerasModel): The KerasModel instance
        """
        dfhistory = pd.DataFrame(model.history)
        epoch = dfhistory['epoch'].iloc[-1]
        if epoch > 0 and epoch % self.save_freq == 0:
            ckpt_path = os.path.join(self.ckpt_dir, f'checkpoint_epoch{epoch}.pt')
            net_dict = model.accelerator.get_state_dict(model.net)
            model.accelerator.save(net_dict, ckpt_path)

            if self.ckpt_list[self.ckpt_idx] != '':
                os.remove(self.ckpt_list[self.ckpt_idx])
            self.ckpt_list[self.ckpt_idx] = ckpt_path
            self.ckpt_idx = (self.ckpt_idx + 1) % self.max_ckpt

    def on_fit_end(self, model: "KerasModel"):
        """Callback at the end of the entire training process

        Args:
            model (KerasModel): The KerasModel instance
        """
        pass



import torch 
from torch import nn
from torch import nn,Tensor 
import torch.nn.functional as F 

from omegaconf import DictConfig
from typing import Dict,Any
from ..base_model import BaseModel

class CatEmbeddingSqrt(nn.Module):
    def __init__(self, categories, d_embed_max = 100):
        super().__init__()
        self.categories = categories
        self.d_embed_list = [min(max(int(x**0.5), 2),d_embed_max) for x in categories]
        self.embedding_list = nn.ModuleList([nn.Embedding(self.categories[i],self.d_embed_list[i])
                            for i in range(len(categories))])
        self.d_cat_sum = sum(self.d_embed_list)
        
    def forward(self, x_cat):
        x_out = torch.cat([self.embedding_list[i](x_cat[:,i]) 
                           for i in range(len(self.categories)) ],dim=1)
        return x_out
    
class MLP(nn.Module):
    def __init__(self, d_in, d_layers, dropout):
        super().__init__()
        layers = []
        for d in d_layers:
            layers.append(nn.Linear(d_in, d))
            layers.append(nn.BatchNorm1d(d))
            layers.append(nn.ReLU())
            layers.append(nn.Dropout(p=dropout))
            d_in = d
        self.mlp = nn.Sequential(*layers)

    def forward(self, x):
        return self.mlp(x)
    
 
class CrossNetVector(nn.Module):
    def __init__(self, d_in, n_cross=2):
        super().__init__()
        self.n_cross = n_cross
        self.linears = nn.ModuleList([nn.Linear(d_in,1,bias=False) for i in range(self.n_cross)])
        self.biases = nn.ParameterList(
            [nn.Parameter(torch.zeros(d_in)) for i in range(self.n_cross)])
        
    def forward(self, x):
        x0 = x
        xi = x
        for i in range(self.n_cross):
            xi = x0*self.linears[i](xi)+self.biases[i]+xi
        return xi
    
    
class CrossNetMatrix(nn.Module):
    def __init__(self, d_in, n_cross=2):
        super().__init__()
        self.n_cross = n_cross
        self.linears = nn.ModuleList([nn.Linear(d_in,d_in) for i in range(self.n_cross)])
        
    def forward(self, x):
        x0 = x
        xi = x
        for i in range(self.n_cross):
            xi = x0*self.linears[i](xi)+xi
        return xi
    

class CrossNetMix(nn.Module):
    def __init__(self, d_in, n_cross =2, low_rank=32, n_experts=4):
        super().__init__()
        self.d_in = d_in
        self.n_cross = n_cross
        self.low_rank = low_rank
        self.n_experts = n_experts

        # U: (d_in, low_rank)
        self.U_list = nn.ParameterList([nn.Parameter(nn.init.xavier_normal_(
            torch.empty(n_experts, d_in, low_rank))) for i in range(self.n_cross)])
        
        # V: (d_in, low_rank)
        self.V_list = nn.ParameterList([nn.Parameter(nn.init.xavier_normal_(
            torch.empty(n_experts, d_in, low_rank))) for i in range(self.n_cross)])
        
        # C: (low_rank, low_rank)
        self.C_list = nn.ParameterList([nn.Parameter(nn.init.xavier_normal_(
            torch.empty(n_experts, low_rank, low_rank))) for i in range(self.n_cross)])
        
        # G: (d_in, 1)
        self.gating = nn.ModuleList([nn.Linear(d_in, 1, bias=False) for i in range(self.n_experts)])

        # Bias 
        self.biases = nn.ParameterList([nn.Parameter(torch.zeros(d_in)) for i in range(self.n_cross)])
        

    def forward(self, x):
        x0 = x
        xi = x
        for i in range(self.n_cross):
            output_of_experts = []
            gating_score_of_experts = []
            for expert_id in range(self.n_experts):
                
                # (1) G(xi)
                # compute the gating score by xi
                gating_score_of_experts.append(self.gating[expert_id](xi))

                # (2) E(xi)
                # project the input xi to low_rank space
                v_x = xi@(self.V_list[i][expert_id])   # (batch_size, low_rank)

                # nonlinear activation in low rank space
                v_x = torch.tanh(v_x)
                v_x = v_x@self.C_list[i][expert_id]     # (batch_size, low_rank)
                v_x = torch.tanh(v_x)

                # project back to d_in space
                uv_x = v_x@(self.U_list[i][expert_id].T)  # (batch_size, d_in)
                expert_out = x0*(uv_x + self.biases[i])
                output_of_experts.append(expert_out)

            # (3) mixture of low-rank experts
            output_of_experts = torch.stack(output_of_experts, 2)  # (batch_size, d_in, n_experts)
            gating_score_of_experts = torch.stack(gating_score_of_experts, 1)  # (batch_size, n_experts, 1)
            moe_out = torch.bmm(output_of_experts, gating_score_of_experts.softmax(1))
            xi = torch.squeeze(moe_out) + xi  # (batch_size, d_in)
            
        return xi
    
    
class DeepCrossBackbone(nn.Module):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__()
        self.hparams = config
        
        d_numerical = self.hparams.continuous_dim
        d_embed_max = self.hparams.input_embed_max

        cross_type = self.hparams.cross_type
        n_cross = self.hparams.cross_order
        
        low_rank = self.hparams.low_rank
        n_experts = self.hparams.experts_num
        
        mlp_layers = [int(x) for x in self.hparams.mlp_layers.split('-')]
        mlp_dropout = self.hparams.mlp_dropout
        stacked = self.hparams.stacked
        
        categories = self.hparams.categorical_cardinality
        n_classes = self.hparams.output_dim 

        if cross_type=='mix':
            assert low_rank is not None and n_experts is not None
            
        self.categories = categories
        self.n_classes = n_classes
        self.stacked = stacked

        self.cat_embedding = CatEmbeddingSqrt(categories, d_embed_max) if categories else None
        
        self.d_in = d_numerical 
        if self.cat_embedding:
            self.d_in+=self.cat_embedding.d_cat_sum
            
        if cross_type=="vector":
            self.cross_layer = CrossNetVector(self.d_in,n_cross)
        elif cross_type=="matrix":
            self.cross_layer = CrossNetMatrix(self.d_in,n_cross)
        elif cross_type=="mix":
            self.cross_layer = CrossNetMix(self.d_in,n_cross,low_rank,n_experts)
        else:
            raise NotImplementedError("cross_type should  be one of ('vector','matrix','mix') !")
        
        self.mlp = MLP(
            d_in= self.d_in,
            d_layers = mlp_layers,
            dropout = mlp_dropout
        )
        
        if self.stacked:
            self.last_linear = nn.Linear(mlp_layers[-1],n_classes)
        else:
            self.last_linear = nn.Linear(self.d_in+mlp_layers[-1],n_classes)
        

    def forward(self, x: Dict):
        
        x_cat,x_num = x["categorical"], x["continuous"]
        
        #embedding 
        x_total = []
        if x_num is not None:
            x_total.append(x_num)
        if self.cat_embedding is not None:
            x_total.append(self.cat_embedding(x_cat))
        x_total = torch.cat(x_total, dim=-1)
        
        
        #cross part
        x_cross = self.cross_layer(x_total)
        
        
        #deep part
        if self.stacked:
            x_deep = self.mlp(x_cross)
            x_out = self.last_linear(x_deep)
        else:
            x_deep = self.mlp(x_total)
            x_deep_cross = torch.cat([x_deep,x_cross],axis = 1)
            x_out = self.last_linear(x_deep_cross)
            
        if self.n_classes==1:
            x_out = x_out.squeeze(-1)
        
        return x_out 


class DeepCrossModel(BaseModel):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__(config, **kwargs)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head
        
    def _build_network(self):
        self._embedding_layer = nn.Identity()
        self._backbone = DeepCrossBackbone(self.hparams)
        setattr(self.backbone, "output_dim", self.hparams.output_dim)
        self._head = nn.Identity()
        
    def forward(self, x: Dict) -> Dict[str, Any]:
        x = self.embed_input(x)
        x = self.compute_backbone(x)
        return self.compute_head(x)

    def extract_embedding(self):
        raise ValueError("Extracting Embeddings is not supported by DeepCrossModel.")




# -*- coding: utf-8 -*-
import smtplib
import traceback 
from email.mime.text import MIMEText

def send_msg(receivers,subject,msg=""):
    
    """
         Examples:
    ---------
    >> subject = "info@train_model.py" #邮件主题
    >> msg = "auc=0.98" #邮件内容
    >> receivers = ["265011xxxx@qq.com"] #收件人
    >> send_msg(receivers,subject,msg)
    """
    
    #设置服务器所需信息
    mail_host = 'smtp.yeah.net'  
    mail_user = 'bugrobot'  
    mail_pass = 'NPWPJBSIVXRTYUOB'   #密码(部分邮箱为授权码) 
    sender = 'bugrobot@yeah.net'  

    #构造邮件内容
    message = MIMEText(msg,'plain','utf-8')  
    message['Subject'] = subject
    message['From'] = sender     
    message['To'] = receivers[0]  

    #登录并发送邮件
    try:
        smtpObj = smtplib.SMTP() 
        #连接到服务器
        smtpObj.connect(mail_host,25)
        #登录到服务器
        smtpObj.login(mail_user,mail_pass) 
        #发送
        smtpObj.sendmail(
            sender,receivers,message.as_string()) 
        #退出
        smtpObj.quit() 
        return 'send_msg success'
    except smtplib.SMTPException as e:
        error = 'send_msg error : '+str(e)
        print(error)
        return error

def monitor_run(function,receivers):
    """
       Examples:
    ---------
    >> receivers = ["265011xxxx@qq.com"] #收件人
    >> def f():
    >>    return 1/0
    >> monitor_run(f,receivers)
    """
    try:
        function()
    except Exception as e:
        error_msg = traceback.format_exc()
        send_msg(receivers,"error@"+function.__name__,error_msg)
        raise e



from dataclasses import dataclass, field
from typing import List, Optional

# from typing import Any, Dict, Iterable, List, Optional


@dataclass
class LinearHeadConfig:
    """A model class for Linear Head configuration; serves as a template and documentation. The models take a
    dictionary as input, but if there are keys which are not present in this model class, it'll throw an exception.

    Args:
        layers (str): Hyphen-separated number of layers and units in the classification/regression head.
                E.g. 32-64-32. Default is just a mapping from intput dimension to output dimension

        activation (str): The activation type in the classification head. The default activation in PyTorch
                like ReLU, TanH, LeakyReLU, etc. https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity

        dropout (float): probability of a classification element to be zeroed.

        use_batch_norm (bool): Flag to include a BatchNorm layer after each Linear Layer+DropOut

        initialization (str): Initialization scheme for the linear layers. Defaults to `kaiming`. Choices
                are: [`kaiming`,`xavier`,`random`].

    """

    layers: str = field(
        default="",
        metadata={
            "help": "Hyphen-separated number of layers and units in the classification/regression head. eg. 32-64-32."
            " Default is just a mapping from intput dimension to output dimension"
        },
    )
    activation: str = field(
        default="ReLU",
        metadata={
            "help": "The activation type in the classification head. The default activation in PyTorch"
            " like ReLU, TanH, LeakyReLU, etc."
            " https://pytorch.org/docs/stable/nn.html#non-linear-activations-weighted-sum-nonlinearity"
        },
    )
    dropout: float = field(
        default=0.0,
        metadata={"help": "probability of an classification element to be zeroed."},
    )
    use_batch_norm: bool = field(
        default=False,
        metadata={"help": "Flag to include a BatchNorm layer after each Linear Layer+DropOut"},
    )
    initialization: str = field(
        default="kaiming",
        metadata={
            "help": "Initialization scheme for the linear layers. Defaults to `kaiming`",
            "choices": ["kaiming", "xavier", "random"],
        },
    )


@dataclass
class MixtureDensityHeadConfig:
    """MixtureDensityHead configuration.

    Args:
        num_gaussian (int): Number of Gaussian Distributions in the mixture model. Defaults to 1

        sigma_bias_flag (bool): Whether to have a bias term in the sigma layer. Defaults to False

        mu_bias_init (Optional[List]): To initialize the bias parameter of the mu layer to predefined
                cluster centers. Should be a list with the same length as number of gaussians in the mixture
                model. It is highly recommended to set the parameter to combat mode collapse. Defaults to None

        weight_regularization (Optional[int]): Whether to apply L1 or L2 Norm to the MDN layers. Defaults
                to L2. Choices are: [`1`,`2`].

        lambda_sigma (Optional[float]): The regularization constant for weight regularization of sigma
                layer. Defaults to 0.1

        lambda_pi (Optional[float]): The regularization constant for weight regularization of pi layer.
                Defaults to 0.1

        lambda_mu (Optional[float]): The regularization constant for weight regularization of mu layer.
                Defaults to 0

        softmax_temperature (Optional[float]): The temperature to be used in the gumbel softmax of the
                mixing coefficients. Values less than one leads to sharper transition between the multiple
                components. Defaults to 1

        n_samples (int): Number of samples to draw from the posterior to get prediction. Defaults to 100

        central_tendency (str): Which measure to use to get the point prediction. Defaults to mean. Choices
                are: [`mean`,`median`].

        speedup_training (bool): Turning on this parameter does away with sampling during training which
                speeds up training, but also doesn't give you visibility on train metrics. Defaults to False

        log_debug_plot (bool): Turning on this parameter plots histograms of the mu, sigma, and pi layers
                in addition to the logits(if log_logits is turned on in experment config). Defaults to False

        input_dim (int): The input dimensions to the head. This will be automatically filled in while
                initializing from the `backbone.output_dim`

    """

    num_gaussian: int = field(
        default=1,
        metadata={
            "help": "Number of Gaussian Distributions in the mixture model. Defaults to 1",
        },
    )
    sigma_bias_flag: bool = field(
        default=False,
        metadata={
            "help": "Whether to have a bias term in the sigma layer. Defaults to False",
        },
    )
    mu_bias_init: Optional[List] = field(
        default=None,
        metadata={
            "help": "To initialize the bias parameter of the mu layer to predefined cluster centers."
            " Should be a list with the same length as number of gaussians in the mixture model."
            " It is highly recommended to set the parameter to combat mode collapse. Defaults to None",
        },
    )

    weight_regularization: Optional[int] = field(
        default=2,
        metadata={
            "help": "Whether to apply L1 or L2 Norm to the MDN layers. Defaults to L2",
            "choices": [1, 2],
        },
    )

    lambda_sigma: Optional[float] = field(
        default=0.1,
        metadata={
            "help": "The regularization constant for weight regularization of sigma layer. Defaults to 0.1",
        },
    )
    lambda_pi: Optional[float] = field(
        default=0.1,
        metadata={
            "help": "The regularization constant for weight regularization of pi layer. Defaults to 0.1",
        },
    )
    lambda_mu: Optional[float] = field(
        default=0,
        metadata={
            "help": "The regularization constant for weight regularization of mu layer. Defaults to 0",
        },
    )
    softmax_temperature: Optional[float] = field(
        default=1,
        metadata={
            "help": "The temperature to be used in the gumbel softmax of the mixing coefficients."
            " Values less than one leads to sharper transition between the multiple components. Defaults to 1",
        },
    )
    n_samples: int = field(
        default=100,
        metadata={
            "help": "Number of samples to draw from the posterior to get prediction. Defaults to 100",
        },
    )
    central_tendency: str = field(
        default="mean",
        metadata={
            "help": "Which measure to use to get the point prediction. Defaults to mean",
            "choices": ["mean", "median"],
        },
    )
    speedup_training: bool = field(
        default=False,
        metadata={
            "help": "Turning on this parameter does away with sampling during training which speeds up training,"
            " but also doesn't give you visibility on train metrics. Defaults to False",
        },
    )
    log_debug_plot: bool = field(
        default=False,
        metadata={
            "help": "Turning on this parameter plots histograms of the mu, sigma, and pi layers in addition"
            " to the logits(if log_logits is turned on in experment config). Defaults to False",
        },
    )
    input_dim: int = field(
        default=None,
        metadata={
            "help": "The input dimensions to the head. This will be automatically filled in while initializing"
            " from the `backbone.output_dim`",
        },
    )
    _probabilistic: bool = field(default=True)


# if __name__ == "__main__":
#     from pytorch_tabular.utils import generate_doc_dataclass
#     print(generate_doc_dataclass(MixtureDensityHeadConfig))



# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
"""Base Model."""
import importlib
import warnings
from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd 
import torch
import torch.nn as nn
from omegaconf import DictConfig, OmegaConf

from torch import Tensor
from torch.optim import Optimizer

from torchkeras.tabular.models.common.heads import blocks
from torchkeras.tabular.models.common.layers import PreEncoded1dLayer
from torchkeras.tabular.utils import reset_all_weights

class BaseModel(nn.Module,metaclass=ABCMeta):
    def __init__(
        self,
        config: DictConfig,
        **kwargs,
    ):
        """Base Model for PyTorch Tabular.

        Args:
            config (DictConfig): The configuration for the model.
            kwargs (Dict, optional): Additional keyword arguments.

        """
        super().__init__()
        self.hparams = config
        self._build_network()
        self._setup_loss()

    @abstractmethod
    def _build_network(self):
        pass

    @property
    def backbone(self):
        raise NotImplementedError("backbone property needs to be implemented by inheriting classes")

    @property
    def embedding_layer(self):
        raise NotImplementedError("embedding_layer property needs to be implemented by inheriting classes")

    @property
    def head(self):
        raise NotImplementedError("head property needs to be implemented by inheriting classes")

    def _get_head_from_config(self):
        _head_callable = getattr(blocks, self.hparams.head)
        return _head_callable(
            in_units=self.backbone.output_dim,
            output_dim=self.hparams.output_dim,
            config=_head_callable._config_template(**self.hparams.head_config),
        )  # output_dim auto-calculated from other configs

    def _setup_loss(self):
        try:
            self.loss = getattr(nn, self.hparams.loss)()
        except AttributeError as e:
            print(f"{self.hparams.loss} is not a valid loss defined in the torch.nn module")
            raise e
       
    def compute_loss(self, output: Dict, y: torch.Tensor) -> torch.Tensor:
        """Calculates the loss for the model.

        Args:
            output (Dict): The output dictionary from the model
            y (torch.Tensor): The target tensor

        Returns:
            torch.Tensor: The loss value

        """
        y_hat = output["logits"]
        reg_terms = [k for k, v in output.items() if "regularization" in k]
        reg_loss = 0
        for t in reg_terms:
            if output[t] != 0:
                reg_loss += output[t]
        if self.hparams.task == "regression":
            computed_loss = reg_loss
            for i in range(self.hparams.output_dim):
                _loss = self.loss(y_hat[:, i], y[:, i])
                computed_loss += _loss
        else:
            # TODO loss fails with batch size of 1?
            computed_loss = self.loss(y_hat.squeeze(), y.squeeze()) + reg_loss
        return computed_loss

    
    def data_aware_initialization(self, dataloader):
        """Performs data-aware initialization of the model before fitted"""
        pass

    def compute_backbone(self, x: Dict) -> torch.Tensor:
        # Returns output
        x = self.backbone(x)
        return x

    def embed_input(self, x: Dict) -> torch.Tensor:
        return self.embedding_layer(x)

    def pack_output(self, y_hat: torch.Tensor, backbone_features: torch.tensor) -> Dict[str, Any]:
        """Packs the output of the model.

        Args:
            y_hat (torch.Tensor): The output of the model

            backbone_features (torch.tensor): The backbone features

        Returns:
            The packed output of the model

        """
        # if self.head is the Identity function it means that we cannot extract backbone features,
        # because the model cannot be divide in backbone and head (i.e. TabNet)
        if type(self.head) is nn.Identity:
            return {"logits": y_hat}
        return {"logits": y_hat, "backbone_features": backbone_features}

    def compute_head(self, backbone_features: Tensor) -> Dict[str, Any]:
        """Computes the head of the model.

        Args:
            backbone_features (Tensor): The backbone features

        Returns:
            The output of the model

        """
        y_hat = self.head(backbone_features)
        return self.pack_output(y_hat, backbone_features)

    def forward(self, x: Dict) -> Dict[str, Any]:
        """The forward pass of the model.

        Args:
            x (Dict): The input of the model with 'continuous' and 'categorical' keys

        """
        x = self.embed_input(x)
        x = self.compute_backbone(x)
        return self.compute_head(x)

    @torch.no_grad()
    def predict(self, x: Dict, ret_model_output: bool = False) -> Union[torch.Tensor, Tuple[torch.Tensor, Dict]]:
        """Predicts the output of the model.

        Args:
            x (Dict): The input of the model with 'continuous' and 'categorical' keys

            ret_model_output (bool): If True, the method returns the output of the model

        Returns:
            The output of the model

        """
        ret_value = self.forward(x)
        if ret_model_output:
            return ret_value.get("logits"), ret_value
        return ret_value.get("logits")

    def extract_embedding(self):
        """Extracts the embedding of the model.

        This is used in `CategoricalEmbeddingTransformer`

        """
        if self.hparams.categorical_dim > 0:
            if not isinstance(self.embedding_layer, PreEncoded1dLayer):
                return self.embedding_layer.cat_embedding_layers
            else:
                raise ValueError(
                    "Cannot extract embedding for PreEncoded1dLayer. Please use a different embedding layer."
                )
        else:
            raise ValueError(
                "Model has been trained with no categorical feature and therefore can't be used"
                " as a Categorical Encoder"
            )

    def reset_weights(self):
        reset_all_weights(self.backbone)
        reset_all_weights(self.head)
        reset_all_weights(self.embedding_layer)

    def feature_importance(self) -> pd.DataFrame:
        """Returns a dataframe with feature importance for the model."""
        if hasattr(self.backbone, "feature_importance_"):
            imp = self.backbone.feature_importance_
            n_feat = len(self.hparams.categorical_cols + self.hparams.continuous_cols)
            if self.hparams.categorical_dim > 0:
                if imp.shape[0] != n_feat:
                    # Combining Cat Embedded Dimensions to a single one by averaging
                    wt = []
                    norm = []
                    ft_idx = 0
                    for _, embd_dim in self.hparams.embedding_dims:
                        wt.extend([ft_idx] * embd_dim)
                        norm.append(embd_dim)
                        ft_idx += 1
                    for _ in self.hparams.continuous_cols:
                        wt.extend([ft_idx])
                        norm.append(1)
                        ft_idx += 1
                    imp = np.bincount(wt, weights=imp) / np.array(norm)
                else:
                    # For models like FTTransformer, we dont need to do anything
                    # It takes categorical and continuous as individual 2-D features
                    pass
            importance_df = pd.DataFrame(
                {
                    "Features": self.hparams.categorical_cols + self.hparams.continuous_cols,
                    "importance": imp,
                }
            )
            return importance_df
        else:
            raise ValueError("Feature Importance unavailable for this model.")




import torch
from transformers import Trainer,TrainingArguments,EarlyStoppingCallback,TrainerCallback
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import is_peft_available
import numpy as np
import pandas as pd 
import os 
import matplotlib.pyplot as plt 
from torchkeras.utils import is_jupyter

"""
@author : lyhue1991
@description : hug model code
"""

class HugModel(torch.nn.Module):
    """Custom PyTorch module for training and evaluating models using the Hugging Face Transformers library."""

    def __init__(self, net=None, loss_fn=None, metrics_dict=None,
                 label_names=None, feature_names=None):
        """
        Args:
            net (torch.nn.Module): The neural network model
            loss_fn: The loss function for training
            metrics_dict (dict): A dictionary of evaluation metrics
            label_names (list): List of label names
            feature_names (list): List of feature names
        """
        super().__init__()
        self.net, self.loss_fn = net, loss_fn
        self.metrics_dict = metrics_dict if metrics_dict is not None else {}

        # Attention here!
        self.feature_names = ['features'] if feature_names is None else feature_names
        self.label_names = ['labels'] if label_names is None else label_names

        self.training_args = None
        self.trainer = None

    def forward(self, **batch):
        """Forward pass of the model.

        Args:
            **batch: Keyword arguments representing the input batch

        Returns:
            dict: Output dictionary containing logits and loss
        """
        # Model forward pass logic
        if self.loss_fn is None:
            return self.net(**batch)
        
        elif set(batch.keys())=={'features','labels'}:
            features = batch['features']
            labels = batch['labels']
            logits = self.net.forward(features)
            loss = self.loss_fn(logits,labels)
            out = {'logits':logits,'loss':loss}
            return out 
        
        else:
            if len(self.feature_names)==1:
                features = batch[self.feature_names[0]]
                logits = self.net.forward(features)
            else:
                features = {k:batch[k] for k in self.feature_names}
                logits = self.net.forward(**features)
            
            if len(self.label_names)==1:
                labels = batch[self.label_names[0]]
            else:
                labels = {k:batch[k] for k in self.label_names}
                
            loss = self.loss_fn(logits,labels)
            out = {'logits':logits,'loss':loss}
            return out 

    def compute_metrics(self, eval_preds):
        """Compute evaluation metrics.

        Args:
            eval_preds: Tuple of logits and labels

        Returns:
            dict: Dictionary of computed metrics
        """
        logits, labels = eval_preds
        for k, m in self.metrics_dict.items():
            m(torch.from_numpy(logits), torch.from_numpy(labels))
        result = {k: m.compute() for k, m in self.metrics_dict.items()}
        for k, m in self.metrics_dict.items():
            m.reset()
        return result

    def get_collate_fn(self, default_fn):
        """Get custom collate function for the data loader

        Args:
            default_fn: Default collate function

        Returns:
            function: Custom collate function
        """
        def collate_fn(examples):
            batch = default_fn(examples)
            if hasattr(batch, 'keys') and hasattr(batch, 'pop'):
                return batch
            elif isinstance(batch, (list, tuple)):
                if len(batch) == 2:
                    return {'features': batch[0], 'labels': batch[1]}
            else:
                raise Exception('dataset format error!')
        return collate_fn

    def fit(self, train_data, val_data=None, output_dir='output_dir',
            epochs=10, learning_rate=5e-5, logging_steps=20,
            monitor="val_loss", patience=20, mode="min",
            plot=True, wandb=False,
            no_cuda=False, use_mps_device=False,
            **kwargs):
        """Train the model.

        Args:
            train_data: Training data
            val_data: Validation data
            output_dir: Output directory for saving checkpoints
            epochs: Number of training epochs
            learning_rate: Learning rate for training
            logging_steps: Number of steps to log information
            monitor: Metric to monitor for early stopping
            patience: Patience for early stopping
            mode: Early stopping mode ('min' or 'max')
            plot: Whether to plot training progress
            wandb: Whether to use Weights & Biases for logging
            no_cuda: Whether to use CUDA
            use_mps_device: Whether to use Mixed Precision Training
            **kwargs: Additional keyword arguments
        """
        self.train_data,self.val_data,self.monitor,self.mode = train_data,val_data,monitor,mode
        
        if wandb ==False:
            os.environ["WANDB_DISABLED"] = "true"
        else:
            os.environ["WANDB_DISABLED"] = "false"
            import wandb as wb
            import datetime
            name =datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            wb.init(
                    project=wandb if isinstance(wandb,str) else "HugModel",
                    name = name,save_code=True
                )
            self.run_id  = wb.run.id
            self.project_name = wb.run.project_name()
            
        self.prefix = 'eval' if monitor.startswith('eval') else 'val'
        monitor_metric = 'e'+monitor if monitor.startswith('val_') else monitor
        
        self.training_args = TrainingArguments(
            output_dir = output_dir,
            num_train_epochs = epochs,
            learning_rate = learning_rate,
            logging_steps = logging_steps,
            
            evaluation_strategy="steps", #epoch
            metric_for_best_model= monitor_metric,
            greater_is_better= False if mode=='min' else True,
            report_to='wandb' if wandb else [],
            
            load_best_model_at_end= True,
            remove_unused_columns = False,
            label_names = self.label_names,
            
            per_device_train_batch_size = train_data.batch_size,
            per_device_eval_batch_size = val_data.batch_size,
            dataloader_drop_last=train_data.drop_last,
            dataloader_num_workers = train_data.num_workers,
            dataloader_pin_memory = train_data.pin_memory,
            
            no_cuda = no_cuda, 
            use_mps_device = use_mps_device,
            **kwargs
        )
        
        callbacks = [EarlyStoppingCallback(early_stopping_patience= patience)]
        
        if plot and is_jupyter():
            callbacks.append(VisCallback(monitor=monitor))
      
        self.trainer = Trainer(
            self,
            self.training_args,
            train_dataset=train_data.dataset,
            eval_dataset=val_data.dataset,
            compute_metrics=self.compute_metrics,
            callbacks = callbacks,
            data_collator=self.get_collate_fn(train_data.collate_fn)
        )
        
        self.trainer.train()
        
        #save best ckpt
        self.ckpt_path = os.path.join(output_dir,'best')
        self.save_ckpt(self.ckpt_path)
                
        if wandb:
            arti_model = wb.Artifact('checkpoint', type='model')
            arti_model.add_file(self.ckpt_path)
            wb.log_artifact(arti_model)
            wb.finish()
    

    def evaluate(self, val_data, **kwargs):
        """Evaluate the model on validation data

        Args:
            val_data: Validation data.
            **kwargs: Additional keyword arguments

        Returns:
            dict: Dictionary of evaluation metrics
        """
        dl_val = self.trainer.get_eval_dataloader(val_data.dataset)
        out = self.trainer.evaluation_loop(dl_val,
                                           description=self.prefix, prediction_loss_only=False,
                                           metric_key_prefix=self.prefix).metrics
        return out

    def save_ckpt(self, ckpt_path):
        """Save model checkpoint

        Args:
            ckpt_path: Path to save the checkpoint
        """
        supported_classes = [PreTrainedModel]
        if is_peft_available():
            from peft import PeftModel
            supported_classes.append(PeftModel)
        supported_classes = tuple(supported_classes)
        if isinstance(self.net, supported_classes) and self.trainer is not None:
            self.trainer.save_model(ckpt_path)
        else:
            torch.save(self.net.state_dict(), ckpt_path)

    def load_ckpt(self, ckpt_path):
        """Load model checkpoint

        Args:
            ckpt_path: Path to load the checkpoint from
        """
        if is_peft_available():
            from peft import PeftModel
            if isinstance(self.net, PeftModel):
                self.net = self.net.from_pretrained(self.net.base_model.model, ckpt_path)
        elif isinstance(self.net, PreTrainedModel):
            self.net = self.net.from_pretrained(ckpt_path)
        else:
            self.net.load_state_dict(torch.load(ckpt_path))
    def debug(self):
        """Debugging method to identify issues during model fitting."""
        # Step 1: Check label_names
        trainer = self.trainer
        print('Step 1: Check label_names.')
        print('label_names = ', trainer.label_names, '\n')

        # Step 2: Check dataset keys
        print('Step 2: Check dataset keys.')
        dl_val = trainer.get_eval_dataloader()
        for step, inputs in enumerate(dl_val):
            break
        print('batch.keys() = ', inputs.keys(), '\n')

        # Step 3: Check labels not None
        print('Step 3: Check labels not None.')
        labels_check = {k + '!=None': inputs.get(k) is not None for k in trainer.label_names}
        print(labels_check)
        has_labels = False if len(trainer.label_names) == 0 else all(
            inputs.get(k) is not None for k in trainer.label_names)
        print('has_labels = ', has_labels, '\n')

        # Step 4: Check trainer.prediction_step
        print('Step 4: Check trainer.prediction_step.')
        loss, logits, labels = trainer.prediction_step(self, inputs, False)
        print('loss = ', loss, '\n')

        # Step 5: Check trainer.evaluation_loop
        print('Step 5: Check trainer.evaluation_loop.')
        out = trainer.evaluation_loop(dl_val, description=self.prefix, metric_key_prefix=self.prefix)
        print('metrics:', out.metrics)

class VisCallback(TrainerCallback):
    def __init__(self, figsize=(6, 4), update_freq=1, monitor=None):
        """
        Visualization callback for Hugging Face Transformers Trainer

        Args:
            figsize (tuple, optional): Figure size. Defaults to (6, 4)
            update_freq (int, optional): Frequency of updating the visualization. Defaults to 1
            monitor (str, optional): Metric to monitor. Defaults to None
        """
        self.figsize = figsize
        self.update_freq = update_freq
        self.monitor = monitor

    def on_train_begin(self, args, state, control, **kwargs):
        """
        Called at the beginning of training

        Args:
            args (TrainingArguments): Training arguments
            state (TrainerState): Trainer state
            control (TrainerControl): Trainer control
        """
        metric = args.metric_for_best_model
        self.greater_is_better = args.greater_is_better
        self.prefix = 'val_' if self.monitor is None or self.monitor.startswith('val_') else 'eval_'
        self.metric = metric if self.monitor is None else self.monitor

        dfhistory = pd.DataFrame()
        x_bounds = [0, args.logging_steps * 10]
        self.update_graph(dfhistory, self.metric.replace(self.prefix, ''), x_bounds=x_bounds, figsize=self.figsize)

    def on_evaluate(self, args, state, control, metrics, **kwargs):
        """
        Called after each evaluation

        Args:
            args (TrainingArguments): Training arguments
            state (TrainerState): Trainer state
            control (TrainerControl): Trainer control
            metrics (Dict): Evaluation metrics
        """
        dfhistory = self.get_history(state)
        n = dfhistory['step'].max()
        if n % self.update_freq == 0:
            x_bounds = [dfhistory['step'].min(),
                        10 * args.logging_steps + (n // (args.logging_steps * 10)) * args.logging_steps * 10]
            self.update_graph(dfhistory, self.metric.replace(self.prefix, ''), x_bounds=x_bounds, figsize=self.figsize)

    def on_train_end(self, args, state, control, **kwargs):
        """
        Called at the end of training

        Args:
            args (TrainingArguments): Training arguments
            state (TrainerState): Trainer state
            control (TrainerControl): Trainer control
        """
        dfhistory = self.get_history(state)
        self.update_graph(dfhistory, self.metric.replace(self.prefix, ''), figsize=self.figsize)

    def get_history(self, state):
        """
        Get training and evaluation history from TrainerState

        Args:
            state (TrainerState): Trainer state

        Returns:
            DataFrame: Combined training and evaluation history
        """
        log_history = state.log_history
        train_history = [x for x in log_history if 'loss' in x.keys()]
        eval_history = [x for x in log_history if 'eval_loss' in x.keys()]

        dfhistory_train = pd.DataFrame(train_history)
        dfhistory_eval = pd.DataFrame(eval_history)
        dfhistory = dfhistory_train.merge(dfhistory_eval, on=['step', 'epoch'])
        if self.prefix == 'val_':
            dfhistory.columns = [x.replace('eval_', 'val_') for x in dfhistory.columns]
        return dfhistory

    def get_best_score(self, dfhistory):
        """
        Get the best score and corresponding step

        Args:
            dfhistory (DataFrame): Combined training and evaluation history

        Returns:
            Tuple: (best_step, best_score)
        """
        arr_scores = dfhistory[self.metric]
        best_score = np.max(arr_scores) if self.greater_is_better else np.min(arr_scores)
        best_step = dfhistory.loc[arr_scores == best_score, 'step'].tolist()[0]
        return best_step, best_score

    def update_graph(self, dfhistory, metric, x_bounds=None, y_bounds=None, figsize=(6, 4)):
        """
        Update the visualization graph

        Args:
            dfhistory (DataFrame): Combined training and evaluation history
            metric (str): Metric to visualize
            x_bounds (list, optional): X-axis bounds. Defaults to None
            y_bounds (list, optional): Y-axis bounds. Defaults to None
            figsize (tuple, optional): Figure size. Defaults to (6, 4)
        """
        from IPython.display import display
        if not hasattr(self, 'graph_fig'):
            self.graph_fig, self.graph_ax = plt.subplots(1, figsize=figsize)
            self.graph_out = display(self.graph_ax.figure, display_id=True)

        self.graph_ax.clear()
        steps = dfhistory['step'] if 'step' in dfhistory.columns else []

        m1 = metric
        if m1 in dfhistory.columns:
            train_metrics = dfhistory[m1]
            self.graph_ax.plot(steps, train_metrics, 'bo--', label=m1, clip_on=False)

        m2 = self.prefix + metric
        if m2 in dfhistory.columns:
            val_metrics = dfhistory[m2]
            self.graph_ax.plot(steps, val_metrics, 'co-', label=m2, clip_on=False)

        self.graph_ax.set_xlabel("step")
        self.graph_ax.set_ylabel(metric)

        if m1 in dfhistory.columns or m2 in dfhistory.columns or metric in dfhistory.columns:
            self.graph_ax.legend(loc='best')

        if len(steps) > 0:
            best_step, best_score = self.get_best_score(dfhistory)
            self.graph_ax.plot(best_step, best_score, 'r*', markersize=15, clip_on=False)
            title = f'best {self.metric} = {best_score:.4f} (@step {best_step})'
            self.graph_ax.set_title(title)
        else:
            title = f'best {self.metric} = ?'
            self.graph_ax.set_title(title)

        if x_bounds is not None:
            self.graph_ax.set_xlim(*x_bounds)
        if y_bounds is not None:
            self.graph_ax.set_ylim(*y_bounds)
        self.graph_out.update(self.graph_ax.figure)
        plt.close()



import gzip
import os
import random
import shutil

import numpy as np
import pandas as pd
import requests
import torch
from pandas import DataFrame, Series
from sklearn.cluster import KMeans
from sklearn.datasets import make_classification, make_regression
from sklearn.preprocessing import LabelEncoder

def _make_smooth_weights_for_balanced_classes(y_train, mu=1.0):
    labels_dict = dict(zip(np.unique(y_train), np.bincount(y_train)))
    total = np.sum(list(labels_dict.values()))
    keys = sorted(labels_dict.keys())
    weight = []
    for i in keys:
        score = np.log(mu * total / float(labels_dict[i]))
        weight.append(score if score > 1 else 1)
    return weight


def get_class_weighted_cross_entropy(y_train, mu=1.0):
    assert y_train.ndim == 1, "Utility function only works for binary classification"
    y_train = LabelEncoder().fit_transform(y_train)
    weights = _make_smooth_weights_for_balanced_classes(y_train, mu=mu)
    criterion = torch.nn.CrossEntropyLoss(weight=torch.FloatTensor(weights))
    return criterion


def get_balanced_sampler(y_train):
    assert y_train.ndim == 1, "Utility function only works for binary classification"
    y_train = LabelEncoder().fit_transform(y_train)
    class_sample_counts = np.bincount(y_train)
    # compute weight for all the samples in the dataset
    # samples_weights contain the probability for each example in dataset to be sampled
    class_weights = 1.0 / torch.Tensor(class_sample_counts)
    train_samples_weight = [class_weights[class_id] for class_id in y_train]
    # now lets initialize samplers
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(train_samples_weight, len(y_train))
    return train_sampler


def get_gaussian_centers(y, n_components):
    if isinstance(y, Series) or isinstance(y, DataFrame):
        y = y.values
    if y.ndim == 1:
        y = y.reshape(-1, 1)
    cluster = KMeans(n_clusters=n_components, random_state=42).fit(y)
    return cluster.cluster_centers_.ravel().tolist()


def make_mixed_dataset(
    task,
    n_samples,
    n_features=7,
    n_categories=2,
    n_informative=5,
    random_state=42,
    n_targets=None,
    **kwargs,
):
    """Creates a synthetic dataset with mixed data types.

    Args:
        task (str): Either "classification" or "regression"
        n_samples (int): Number of samples to generate
        n_features (int): Number of total features to generate
        n_categories (int): Number of features to be categorical
        n_informative (int): Number of informative features
        random_state (int): Random seed for reproducibility
        n_targets (int): Number of targets to generate. n_targets>1 will generate a multi-target dataset
            for regression and multi-class dataset for classification.
            Defaults to 2 classes for classification and 1 for regression
        kwargs: Additional arguments to pass to the make_classification or make_regression function

    """
    assert n_features >= n_categories, "n_features must be greater than or equal to n_categories"
    assert n_informative <= n_features, "n_informative must be less than or equal to n_features"
    assert task in [
        "classification",
        "regression",
    ], "task must be either classification or regression"
    if n_targets is None:
        n_targets = 1 if task == "regression" else 2
    if task == "classification":
        X, y = make_classification(
            n_samples=n_samples,
            n_features=n_features,
            random_state=random_state,
            n_informative=n_informative,
            n_classes=n_targets,
            **kwargs,
        )
    elif task == "regression":
        X, y = make_regression(
            n_samples=n_samples,
            n_features=n_features,
            random_state=random_state,
            n_informative=n_informative,
            n_targets=n_targets,
            **kwargs,
        )
    cat_cols = random.choices(list(range(X.shape[-1])), k=n_categories)
    num_cols = [i for i in range(X.shape[-1]) if i not in cat_cols]
    for col in cat_cols:
        X[:, col] = pd.qcut(X[:, col], q=4).codes.astype(int)
    col_names = []
    num_col_names = []
    cat_col_names = []
    for i in range(X.shape[-1]):
        if i in cat_cols:
            col_names.append(f"cat_col_{i}")
            cat_col_names.append(f"cat_col_{i}")
        if i in num_cols:
            col_names.append(f"num_col_{i}")
            num_col_names.append(f"num_col_{i}")
    X = pd.DataFrame(X, columns=col_names)
    if n_targets == 1 or task == "classification":
        y = pd.Series(y, name="target")
    else:
        y = pd.DataFrame(y, columns=[f"target_{i}" for i in range(n_targets)])
    if task == "classification":
        y = "class_" + y.astype(str)
    data = X.join(y)
    return data, cat_col_names, num_col_names


def print_metrics(metrics, y_true, y_pred, tag, return_dict=False):
    if isinstance(y_true, pd.DataFrame) or isinstance(y_true, pd.Series):
        y_true = y_true.values
    if isinstance(y_pred, pd.DataFrame) or isinstance(y_pred, pd.Series):
        y_pred = y_pred.values
    if y_true.ndim > 1:
        y_true = y_true.ravel()
    if y_pred.ndim > 1:
        y_pred = y_pred.ravel()
    print_str_l = []
    res_d = {}
    for metric, name, params in metrics:
        score = metric(y_true, y_pred, **params)
        print_str_l.append(f"{tag} {name}: {score}")
        res_d[name] = score
    print((" | ".join(print_str_l)).strip())
    if return_dict:
        return res_d


def load_covertype_dataset(download_dir=None):
    """Predicting forest cover type from cartographic variables only (no remotely sensed data). The actual forest cover
    type for a given observation (30 x 30 meter cell) was determined from US Forest Service (USFS) Region 2 Resource
    Information System (RIS) data. Independent variables were derived from data originally obtained from US Geological
    Survey (USGS) and USFS data. Data is in raw form (not scaled) and contains binary (0 or 1) columns of data for
    qualitative independent variables (wilderness areas and soil types).

    This study area includes four wilderness areas located in the Roosevelt National Forest of northern Colorado.
    These areas represent forests with minimal human-caused disturbances, so that existing forest cover types are more a
    result of ecological processes rather than forest management practices.

    It is from [UCI ML Repository](https://archive.ics.uci.edu/ml/datasets/covertype), but with small changes:
    - The one hot encoded columns are converted to categorical - Soli Type and Wilderness type

    Args:
        download_dir (str): Directory to download the data to. Defaults to None, which will download
            to ~/.pytorch_tabular/datasets/

    """
    if download_dir is None:
        download_dir = os.path.join(os.path.expanduser("~"), ".pytorch_tabular", "datasets")
    if not os.path.exists(download_dir):
        os.makedirs(download_dir)
    file_path = os.path.join(download_dir, "covertype.csv")
    if not os.path.exists(file_path):
        logger.info("Downloading Covertype Dataset")
        url = "https://archive.ics.uci.edu/ml/machine-learning-databases/covtype/covtype.data.gz"
        response = requests.get(url)
        with open(os.path.join(download_dir, "covertype.data.gz"), "wb") as f:
            f.write(response.content)
        with gzip.open(os.path.join(download_dir, "covertype.data.gz"), "rb") as f_in:
            with open(os.path.join(download_dir, "covertype.csv"), "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(os.path.join(download_dir, "covertype.data.gz"))
    df = pd.read_csv(file_path, header=None)
    df.columns = (
        [
            "Elevation",
            "Aspect",
            "Slope",
            "Horizontal_Distance_To_Hydrology",
            "Vertical_Distance_To_Hydrology",
            "Horizontal_Distance_To_Roadways",
            "Hillshade_9am",
            "Hillshade_Noon",
            "Hillshade_3pm",
            "Horizontal_Distance_To_Fire_Points",
        ]
        + [f"Wilderness_Area_{i}" for i in range(4)]
        + [f"Soil_Type_{i}" for i in range(40)]
        + ["Cover_Type"]
    )
    # convert one hot encoded columns to categorical
    df["Wilderness_Area"] = df[[f"Wilderness_Area_{i}" for i in range(4)]].idxmax(axis=1).str.split("_").str[-1]
    df["Soil_Type"] = df[[f"Soil_Type_{i}" for i in range(40)]].idxmax(axis=1).str.split("_").str[-1]
    df.drop(
        [f"Wilderness_Area_{i}" for i in range(4)] + [f"Soil_Type_{i}" for i in range(40)],
        axis=1,
        inplace=True,
    )
    continuous_cols = [
        "Elevation",
        "Aspect",
        "Slope",
        "Horizontal_Distance_To_Hydrology",
        "Vertical_Distance_To_Hydrology",
        "Horizontal_Distance_To_Roadways",
        "Hillshade_9am",
        "Hillshade_Noon",
        "Hillshade_3pm",
        "Horizontal_Distance_To_Fire_Points",
    ]
    categorical_cols = ["Wilderness_Area", "Soil_Type"]
    return df, categorical_cols, continuous_cols, "Cover_Type"



import pandas as pd
from tqdm import tqdm
import torch
from .utils import printlog, prettydf

"""
@author : lyhue1991
@description : keras soup code
"""

def merge_state(state_list, weights=None):
    """
    Merges states from multiple checkpoints with optional weights

    Args:
        state_list (list): List of checkpoint paths or state dictionaries
        weights (list): List of weights corresponding to each state. If None, equal weights are assigned

    Returns:
        dict: Merged state dictionary
    """
    if weights is None:
        weights = [1 for _ in state_list]
    total = sum(weights)
    weights = [x / total for x in weights]

    if isinstance(state_list[0], str):
        state = {k: v * weights[0] for k, v in torch.load(state_list[0]).items()}
    else:
        state = {k: v * weights[0] for k, v in state_list[0].items()}

    for st, w in zip(state_list[1:], weights[1:]):
        if isinstance(st, str):
            st = torch.load(st)
        for k, v in st.items():
            state[k] = state[k] + w * v

    return state

def uniform_soup(model, ckpt_path_list, saved_ckpt_path='checkpoint_uniform_soup.pt'):
    """
    Uniformly merges states from multiple checkpoints and evaluates the model

    Args:
        model: The PyTorch model
        ckpt_path_list (list): List of checkpoint paths
        saved_ckpt_path (str): Path to save the merged checkpoint

    Returns:
        float: Evaluation score
    """
    state = merge_state(ckpt_path_list)
    model.net.load_state_dict(state)
    score = model.evaluate(model.val_data, quiet=True)[model.monitor]
    torch.save(model.net.state_dict(), saved_ckpt_path)
    return score

def greedy_soup(model, ckpt_path_list, num_models=120, num_warmup=0, saved_ckpt_path='checkpoint_greedy_soup.pt'):
    """
    Greedily merges states from multiple checkpoints and evaluates the model

    Args:
        model: The PyTorch model
        ckpt_path_list (list): List of checkpoint paths
        num_models (int): Number of models to merge
        num_warmup (int): Number of warmup models (do not choose greedily)
        saved_ckpt_path (str): Path to save the merged checkpoint

    Returns:
        float: Evaluation score
    """
    dfckpt = pd.DataFrame({'ckpt_path':ckpt_path_list})

    scores = []
    printlog('step1: sort ckpt_path by metric...')
    
    loop = tqdm(dfckpt['ckpt_path'])
    for ckpt_path in loop:
        model.load_ckpt(ckpt_path)
        score = model.evaluate(model.val_data,quiet=True)[model.monitor]
        scores.append(score)
        loop.set_postfix(**{model.monitor:score})

    dfckpt['score'] = scores
    dfckpt = dfckpt.sort_values('score',ascending= True if model.mode=='min' else False) 
    prettydf(dfckpt,str_len=50,show=True)
    
    dfckpt['weights'] = [0 for _ in dfckpt['ckpt_path']]
    
    loop = tqdm(range(num_models),total = num_models)
    
    printlog('step2: start greedy merge...')
    
    for i in loop:
        
        if i==0:
            cur_state = torch.load(dfckpt['ckpt_path'].iloc[0])
            cur_score = dfckpt['score'].iloc[0]
            dfckpt['weights'].iloc[0] = 1
            loop.set_postfix(**{'i':i,'score':cur_score})

        elif i+1<=num_warmup:
            state_i = torch.load(dfckpt['ckpt_path'].iloc[i])
            cur_state = merge_state(state_list = [cur_state, state_i],
                                    weights =[i,1])
            dfckpt['weights'].iloc[i] = dfckpt['weights'].iloc[i]+1 
            model.net.load_state_dict(cur_state)
            cur_score = model.evaluate(model.val_data,quiet=True)[model.monitor]
            
        else:
            for j in range(len(dfckpt)):
                state_j = torch.load(dfckpt['ckpt_path'].iloc[j])
                maybe_state = merge_state(state_list = [cur_state,state_j],
                                          weights = [i,1]
                                         )
                model.net.load_state_dict(maybe_state)
                maybe_score = model.evaluate(model.val_data,quiet=True)[model.monitor]
                if (model.mode=='max' and maybe_score>cur_score) or (
                    model.mode=='min' and maybe_score<cur_score):
                    
                    cur_state = maybe_state
                    cur_score = maybe_score
                    dfckpt['weights'].iloc[j] = dfckpt['weights'].iloc[j]+1
                    loop.set_postfix(**{'i':i,'score':cur_score})
                    break
            else:
                loop.set_postfix(**{'i':i,'score':cur_score})
                loop.close()
                print('could not get better score, early stopping...')
                break 
 
    printlog('step3: save result...')
    model.net.load_state_dict(cur_state)
    torch.save(model.net.state_dict(),saved_ckpt_path)
    prettydf(dfckpt,str_len=50,show=True)
    print('best_score = ', cur_score)
    print('greedy soup ckpt saved at path: '+saved_ckpt_path)
    
    return cur_score


def optuna_soup(model,
                 ckpt_path_list,
                 n_trials=50,
                 timeout=1200,
                 saved_ckpt_path='checkpoint_optuna_soup.pt',
                 plot=True):
    """
    Perform an Optuna search to find optimal weights for checkpoint ensemble

    Args:
        model: The PyTorch model
        ckpt_path_list (list): List of checkpoint paths
        n_trials (int): Number of optimization trials
        timeout (int): Timeout for the optimization in seconds
        saved_ckpt_path (str): Path to save the final checkpoint
        plot (bool): Whether to plot the optimization analysis

    Returns:
        float: Best evaluation score
    """
    import optuna
    optuna.logging.set_verbosity(optuna.logging.WARNING)

    def objective(trial):
        weights_dict = {name: trial.suggest_int(name, 1, 100) for name in ckpt_path_list}
        state = merge_state(ckpt_path_list, weights=[weights_dict[name] for name in ckpt_path_list])
        model.net.load_state_dict(state)
        score = model.evaluate(model.val_data, quiet=True)[model.monitor]
        print(score)
        return score

    study = optuna.create_study(
        direction="maximize" if model.mode == 'max' else "minimize",
        study_name="optuna_ensemble"
    )

    printlog('step1: start Optuna search...')
    study.optimize(objective, n_trials=n_trials, timeout=timeout)

    printlog('step2: save result...')
    best_weights = study.best_params
    best_score = study.best_value

    state = merge_state(ckpt_path_list, weights=[best_weights[name] for name in ckpt_path_list])
    model.net.load_state_dict(state)
    torch.save(model.net.state_dict(), saved_ckpt_path)

    print(f"best_score = {best_score}")
    print("best_weights:")
    print(best_weights)
    print(f'Optuna soup ckpt saved at path: {saved_ckpt_path}')

    if plot:
        printlog('step3: plot Optuna analysis...')
        optuna.visualization.plot_optimization_history(study).show()
        optuna.visualization.plot_parallel_coordinate(study).show()

    return best_score



import random
from typing import Callable
from warnings import warn

import numpy as np
import torch
import torch.nn as nn

from torchkeras.tabular.utils import check_numpy

from ..layers import ModuleWithInit
from .activations import RSoftmax, entmax15, sparsemax, sparsemoid

# Neural Oblivious Decision Ensembles
# Author: Sergey Popov, Julian Qian
# https://github.com/Qwicen/node
# For license information, see https://github.com/Qwicen/node/blob/master/LICENSE.md


class ODST(ModuleWithInit):
    def __init__(
        self,
        in_features,
        num_trees,
        depth=6,
        tree_output_dim=1,
        flatten_output=True,
        choice_function=sparsemax,
        bin_function=sparsemoid,
        initialize_response_=nn.init.normal_,
        initialize_selection_logits_=nn.init.uniform_,
        threshold_init_beta=1.0,
        threshold_init_cutoff=1.0,
    ):
        """Oblivious Differentiable Sparsemax Trees. http://tinyurl.com/odst-readmore One can drop (sic!) this module
        anywhere instead of nn.Linear.

        :param in_features: number of features in the input tensor
        :param num_trees: number of trees in this layer
        :param tree_dim: number of response channels in the response of individual tree
        :param depth: number of splits in every tree
        :param flatten_output: if False, returns [..., num_trees, tree_dim],
            by default returns [..., num_trees * tree_dim]
        :param choice_function: f(tensor, dim) -> R_simplex computes feature weights s.t. f(tensor, dim).sum(dim) == 1
        :param bin_function: f(tensor) -> R[0, 1], computes tree leaf weights

        :param initialize_response_: in-place initializer for tree output tensor
        :param initialize_selection_logits_: in-place initializer for logits that select features for the tree
        both thresholds and scales are initialized with data-aware init (or .load_state_dict)
        :param threshold_init_beta: initializes threshold to a q-th quantile of data points
            where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
            If this param is set to 1, initial thresholds will have the same distribution as data points
            If greater than 1 (e.g. 10), thresholds will be closer to median data value
            If less than 1 (e.g. 0.1), thresholds will approach min/max data values.

        :param threshold_init_cutoff: threshold log-temperatures initializer, in (0, inf)
            By default(1.0), log-remperatures are initialized in such a way that all bin selectors
            end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
            Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
            Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
            For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
            Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
            All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
        """
        super().__init__()
        self.depth, self.num_trees, self.tree_dim, self.flatten_output = (
            depth,
            num_trees,
            tree_output_dim,
            flatten_output,
        )
        self.choice_function, self.bin_function = choice_function, bin_function
        self.threshold_init_beta, self.threshold_init_cutoff = (
            threshold_init_beta,
            threshold_init_cutoff,
        )

        self.response = nn.Parameter(torch.zeros([num_trees, tree_output_dim, 2**depth]), requires_grad=True)
        initialize_response_(self.response)

        self.feature_selection_logits = nn.Parameter(torch.zeros([in_features, num_trees, depth]), requires_grad=True)
        initialize_selection_logits_(self.feature_selection_logits)

        self.feature_thresholds = nn.Parameter(
            torch.full([num_trees, depth], float("nan"), dtype=torch.float32),
            requires_grad=True,
        )  # nan values will be initialized on first batch (data-aware init)
        self.log_temperatures = nn.Parameter(
            torch.full([num_trees, depth], float("nan"), dtype=torch.float32),
            requires_grad=True,
        )

        # binary codes for mapping between 1-hot vectors and bin indices
        with torch.no_grad():
            indices = torch.arange(2**self.depth)
            offsets = 2 ** torch.arange(self.depth)
            bin_codes = (indices.view(1, -1) // offsets.view(-1, 1) % 2).to(torch.float32)
            bin_codes_1hot = torch.stack([bin_codes, 1.0 - bin_codes], dim=-1)
            self.bin_codes_1hot = nn.Parameter(bin_codes_1hot, requires_grad=False)
            # ^-- [depth, 2 ** depth, 2]

    def forward(self, input):
        assert len(input.shape) >= 2
        if len(input.shape) > 2:
            return self.forward(input.view(-1, input.shape[-1])).view(*input.shape[:-1], -1)
        # new input shape: [batch_size, in_features]

        feature_logits = self.feature_selection_logits
        feature_selectors = self.choice_function(feature_logits, dim=0)
        # ^--[in_features, num_trees, depth]

        feature_values = torch.einsum("bi,ind->bnd", input, feature_selectors)
        # ^--[batch_size, num_trees, depth]

        threshold_logits = (feature_values - self.feature_thresholds) * torch.exp(-self.log_temperatures)

        threshold_logits = torch.stack([-threshold_logits, threshold_logits], dim=-1)
        # ^--[batch_size, num_trees, depth, 2]

        bins = self.bin_function(threshold_logits)
        # ^--[batch_size, num_trees, depth, 2], approximately binary

        bin_matches = torch.einsum("btds,dcs->btdc", bins, self.bin_codes_1hot)
        # ^--[batch_size, num_trees, depth, 2 ** depth]

        response_weights = torch.prod(bin_matches, dim=-2)
        # ^-- [batch_size, num_trees, 2 ** depth]

        response = torch.einsum("bnd,ncd->bnc", response_weights, self.response)
        # ^-- [batch_size, num_trees, tree_dim]

        return response.flatten(1, 2) if self.flatten_output else response

    def initialize(self, input, eps=1e-6):
        # data-aware initializer
        assert len(input.shape) == 2
        if input.shape[0] < 1000:
            warn(
                "Data-aware initialization is performed on less than 1000 data points. This may cause instability."
                "To avoid potential problems, run this model on a data batch with at least 1000 data samples."
                "You can do so manually before training. Use with torch.no_grad() for memory efficiency."
            )
        with torch.no_grad():
            feature_selectors = self.choice_function(self.feature_selection_logits, dim=0)
            # ^--[in_features, num_trees, depth]

            feature_values = torch.einsum("bi,ind->bnd", input, feature_selectors)
            # ^--[batch_size, num_trees, depth]

            # initialize thresholds: sample random percentiles of data
            percentiles_q = 100 * np.random.beta(
                self.threshold_init_beta,
                self.threshold_init_beta,
                size=[self.num_trees, self.depth],
            )
            self.feature_thresholds.data[...] = torch.as_tensor(
                list(
                    map(
                        np.percentile,
                        check_numpy(feature_values.flatten(1, 2).t()),
                        percentiles_q.flatten(),
                    )
                ),
                dtype=feature_values.dtype,
                device=feature_values.device,
            ).view(self.num_trees, self.depth)

            # init temperatures: make sure enough data points are in the linear region of sparse-sigmoid
            temperatures = np.percentile(
                check_numpy(abs(feature_values - self.feature_thresholds)),
                q=100 * min(1.0, self.threshold_init_cutoff),
                axis=0,
            )

            # if threshold_init_cutoff > 1, scale everything down by it
            temperatures /= max(1.0, self.threshold_init_cutoff)
            self.log_temperatures.data[...] = torch.log(torch.as_tensor(temperatures) + eps)

    def __repr__(self):
        return "{}(in_features={}, num_trees={}, depth={}, tree_dim={}, flatten_output={})".format(
            self.__class__.__name__,
            self.feature_selection_logits.shape[0],
            self.num_trees,
            self.depth,
            self.tree_dim,
            self.flatten_output,
        )


class NeuralDecisionStump(nn.Module):
    def __init__(
        self,
        n_features: int,
        binning_activation: Callable = entmax15,
        feature_mask_function: Callable = entmax15,
        feature_sparsity: float = 0.8,
        learnable_sparsity: bool = True,
    ):
        super().__init__()
        self._num_cutpoints = 1
        self._num_leaf = 2
        self.n_features = n_features
        self.binning_activation = binning_activation
        self.feature_mask_function = feature_mask_function
        self.feature_sparsity = feature_sparsity
        self.learnable_sparsity = learnable_sparsity
        self._build_network()

    def _build_network(self):
        # sampling a random beta distribution
        # random distribution helps with diversity in trees and feature splits
        alpha = random.uniform(0.5, 10.0)
        beta = random.uniform(0.5, 10.0)
        # with torch.no_grad():
        feature_mask = (
            torch.distributions.Beta(torch.tensor([alpha]), torch.tensor([beta])).sample((self.n_features,)).squeeze(-1)
        )
        self.feature_mask = nn.Parameter(feature_mask, requires_grad=True)
        if self.feature_mask_function.__name__ == "t_softmax":
            t = RSoftmax.calculate_t(self.feature_mask, r=torch.tensor([self.feature_sparsity]))
            self.t = nn.Parameter(t, requires_grad=self.learnable_sparsity)
        W = torch.linspace(
            1.0,
            self._num_cutpoints + 1.0,
            self._num_cutpoints + 1,
            requires_grad=False,
        ).reshape(1, 1, -1)
        self.register_buffer("W", W)

        cutpoints = torch.rand([self.n_features, self._num_cutpoints])
        # Append zeros to the beginning of each row
        cutpoints = torch.cat([torch.zeros([self.n_features, 1], device=cutpoints.device), cutpoints], 1)
        self.cut_points = nn.Parameter(cutpoints, requires_grad=True)
        self.leaf_responses = nn.Parameter(torch.rand(self.n_features, self._num_leaf), requires_grad=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.feature_mask_function.__name__ == "t_softmax":
            t = torch.relu(self.t)
            feature_mask = self.feature_mask_function(self.feature_mask, t)
        else:
            feature_mask = self.feature_mask_function(self.feature_mask)
        # Repeat W for each batch size using broadcasting
        W = torch.ones(x.size(0), 1, 1, device=x.device) * self.W
        # Binning features
        x = torch.bmm(x.unsqueeze(-1), W) - self.cut_points.unsqueeze(0)
        x = self.binning_activation(x)  # , dim=-1)
        x = x * self.leaf_responses.unsqueeze(0)
        x = (x * feature_mask.reshape(1, -1, 1)).sum(dim=1)
        return x, feature_mask


class NeuralDecisionTree(nn.Module):
    def __init__(
        self,
        depth: int,
        n_features: int,
        dropout: float = 0,
        binning_activation: Callable = entmax15,
        feature_mask_function: Callable = entmax15,
        feature_sparsity: float = 0.8,
        learnable_sparsity: bool = True,
    ):
        super().__init__()
        self.depth = depth
        self._num_cutpoints = 1
        self.n_features = n_features
        self._dropout = dropout
        self.binning_activation = binning_activation
        self.feature_mask_function = feature_mask_function
        self.feature_sparsity = feature_sparsity
        self.learnable_sparsity = learnable_sparsity
        self._build_network()

    def _build_network(self):
        for d in range(self.depth):
            for n in range(max(2 ** (d), 1)):
                self.add_module(
                    f"decision_stump_{d}_{n}",
                    NeuralDecisionStump(
                        self.n_features + (2 ** (d) if d > 0 else 0),
                        self.binning_activation,
                        self.feature_mask_function,
                        self.feature_sparsity,
                        self.learnable_sparsity,
                    ),
                )
        self.dropout = nn.Dropout(self._dropout)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        tree_input = x
        feature_masks = []
        for d in range(self.depth):
            layer_nodes = []
            layer_feature_masks = []
            for n in range(max(2 ** (d), 1)):
                leaf_nodes, feature_mask = self._modules[f"decision_stump_{d}_{n}"](tree_input)
                layer_nodes.append(leaf_nodes)
                layer_feature_masks.append(feature_mask)
            layer_nodes = torch.cat(layer_nodes, dim=1)
            tree_input = torch.cat([x, layer_nodes], dim=1)
            feature_masks.append(layer_feature_masks)
        return self.dropout(layer_nodes), feature_masks



import torch
import sys 
from copy import deepcopy
from .conversations import conv_templates, get_conv_template 
from .text2ids import build_inputs_labels

#chat tool for chatglm2-6b,baichuan-13b,internlm-chat-7b,qwen-7b-chat and more...
class ChatLLM:
    def __init__(self,model,tokenizer,
                 model_type=None,
                 max_chat_rounds=20,
                 max_new_tokens=512,
                 stream=True,
                 history=None,
                 stop_words_ids=None
                ):
        self.model = model
        self.tokenizer = tokenizer
        
        if not self.tokenizer.eos_token_id:
            self.tokenizer.eos_token_id =  (model.config.eos_token_id 
                or model.generation_config.eos_token_id)
            
        self.model_type = model_type if model_type else self.get_model_type() 
        conv = get_conv_template(self.model_type)
        self.conv_template = conv

        stop_words_ids = stop_words_ids if stop_words_ids else []
        stop_token_ids = [[w] for w in conv.stop_token_ids] if conv.stop_token_ids else []
        self.stop_words_ids = stop_token_ids + stop_words_ids
        
        self.model.generation_config.stop_words_ids = self.stop_words_ids
        self.model.generation_config.max_new_tokens = max_new_tokens
        self.model.eval()
        self.history = [] if history is None else history
        self.max_chat_rounds = max_chat_rounds
        self.stream = stream
        
        try:
            self.register_magic() 
            response = self('你好')
            if not self.stream:
                print(response)
            print('register magic %%chat sucessed ...',file=sys.stderr)
            self.history = self.history[:-1]
        except Exception as err:
            print('register magic %%chat failed ...',file=sys.stderr)
            
        
    def get_model_type(self):
        model_cls = str(self.model.__class__).split('.')[-1].lower()[:-2] 
        keys = list(conv_templates.keys()) 
        max_same,most_type = 0,None
        for k in keys:
            same = 0
            for a,b in zip(k,model_cls):
                if a==b:
                    same+=1
                else:
                    break
            if same>max_same:
                max_same = same
                most_type = k 
        if max_same>=3:
            return most_type
        else:
            raise Exception('Error: get_model_type failed @ model_cls='+model_cls)
            return None
        
    @classmethod
    def build_messages(cls,query=None,history=None,system=None):
        messages = []
        history = history if history else [] 
        if system is not None:
            messages.append({'role':'system','content':system})
        for prompt,response in history:
            pair = [{"role": "user", "content": prompt},
                {"role": "assistant", "content": response}]
            messages.extend(pair)
        if query is not None:
            messages.append({"role": "user", "content": query})
        return messages

    def build_conversations(self,messages):
        conv = deepcopy(self.conv_template)
        msgs_sys = [d for d in messages if d['role']=='system']
        if msgs_sys:
            conv.set_system_message(msgs_sys[0]['content'])
        for d in messages:
            if d['role']=='user':
                conv.append_message(conv.roles[0], d['content'])
            elif d['role']=='assistant':
                conv.append_message(conv.roles[1], d['content'])
            else:
                raise Exception('role must be one of (system,user,assistant)')
        if d['role']!='assistant':
            conv.append_message(conv.roles[1], None)
        return conv
        
    def build_prompt(self,messages):
        conv = self.build_conversations(messages)
        return conv.get_prompt()

    def build_inputs_labels(self,messages,multi_rounds=True):
        conv = self.build_conversations(messages)
        inputs,labels = build_inputs_labels(
            conv, self.tokenizer, multi_rounds=multi_rounds)
        return inputs,labels
        
    def chat(self, messages, stream=False, generation_config=None):
        model,tokenizer = self.model,self.tokenizer
        #prompt = self.build_prompt(messages)
        #inputs = tokenizer([prompt],add_special_tokens=False)
        input_ids,labels = self.build_inputs_labels(messages)
        inputs = {'input_ids': torch.tensor([input_ids]).to(model.device)}
        if generation_config is not None:
            generation_config = deepcopy(model.generation_config.update(**generation_config))
        else:
            generation_config =  deepcopy(model.generation_config)

        stop_words_ids = self.stop_words_ids

        if not stream:
            from transformers import PreTrainedModel
            model.__class__.generate = PreTrainedModel.generate  # disable stream
            output_ids = model.generate(
                 **inputs,
                 stop_words_ids = stop_words_ids,
                 generation_config = generation_config,
                 return_dict_in_generate = False,
            )
            output_ids = output_ids[0][len(inputs["input_ids"][0]):]
            end_token_idx = 0
            should_stop = False
            for end_token_idx in range(len(output_ids)):
                for stop_ids in stop_words_ids:
                    if output_ids[end_token_idx-len(stop_ids):end_token_idx].tolist()==stop_ids:
                        should_stop = True
                        break
                if should_stop:
                    break      
            outputs = tokenizer.decode(
                output_ids[:end_token_idx], skip_special_tokens=True
            )
            return outputs
        else:
            from .stream_generate  import NewGenerationMixin, StreamGenerationConfig
            model.__class__.generate = NewGenerationMixin.generate
            model.__class__.sample_stream = NewGenerationMixin.sample_stream
            config_dic = generation_config.to_dict()
            config_dic.update({'do_stream':True})
            stream_config = StreamGenerationConfig(**config_dic)
            
            def stream_generator():
                outputs = []
                for token in model.generate(**inputs,
                                            generation_config=stream_config,
                                            do_sample=True,
                                            stop_words_ids=stop_words_ids,
                                            return_dict_in_generate = False,
                                           ):
                    token_idx = token.item()
                    outputs.append(token_idx)
                    should_stop = False
                    for stop_ids in stop_words_ids:
                        if outputs[-len(stop_ids):]==stop_ids:
                            should_stop = True
                            break
                    if should_stop:
                        break
                    yield tokenizer.decode(outputs, skip_special_tokens=True)
            return stream_generator()
        
    def __call__(self,query):
        from IPython.display import display,clear_output 
        len_his = len(self.history)
        if len_his>=self.max_chat_rounds+1:
            self.history = self.history[len_his-self.max_chat_rounds:]
        messages = self.build_messages(query=query,history=self.history)
        if not self.stream:
            response = self.chat(messages,stream=False)
            self.history.append((query,response))
            return response 
        
        result = self.chat(messages,stream=True)
        for response in result:
            print(response)
            clear_output(wait=True)
        self.history.append((query,response))
        return response
    
    def register_magic(self):
        import IPython
        from IPython.core.magic import (Magics, magics_class, line_magic,
                                        cell_magic, line_cell_magic)
        @magics_class
        class ChatMagics(Magics):
            def __init__(self,shell, pipe):
                super().__init__(shell)
                self.pipe = pipe

            @line_cell_magic
            def chat(self, line, cell=None):
                "Magic that works both as %chat and as %%chat"
                if cell is None:
                    return self.pipe(line)
                else:
                    print(self.pipe(cell))       
        ipython = IPython.get_ipython()
        magic = ChatMagics(ipython,self)
        ipython.register_magics(magic)


# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
# Inspired by implementations
# 1. lucidrains - https://github.com/lucidrains/tab-transformer-pytorch/
# If you are interested in Transformers, you should definitely check out his repositories.
# 2. PyTorch Wide and Deep - https://github.com/jrzaurin/pytorch-widedeep/
# It is another library for tabular data, which supports multi modal problems.
# Check out the library if you haven't already.
# 3. AutoGluon - https://github.com/awslabs/autogluon
# AutoGluon is an AuttoML library which supports Tabular data as well. it is from Amazon Research and is in MXNet
# 4. LabML Annotated Deep Learning Papers - The position-wise FF was shamelessly copied from
# https://github.com/labmlai/annotated_deep_learning_paper_implementations/tree/master/labml_nn/transformers
"""TabTransformer Model."""
from collections import OrderedDict
from typing import Dict

import torch
import torch.nn as nn
from einops import rearrange
from omegaconf import DictConfig

from ..common.layers.batch_norm import BatchNorm1d
from ..base_model import BaseModel
from ..common.layers import Embedding2dLayer, TransformerEncoderBlock


class TabTransformerBackbone(nn.Module):
    def __init__(self, config: DictConfig):
        super().__init__()
        assert config.share_embedding_strategy in [
            "add",
            "fraction",
        ], (
            f"`share_embedding_strategy` should be one of `add` or `fraction`,"
            f" not {self.hparams.share_embedding_strategy}"
        )
        self.hparams = config
        self._build_network()

    def _build_network(self):
        self.transformer_blocks = OrderedDict()
        for i in range(self.hparams.num_attn_blocks):
            self.transformer_blocks[f"mha_block_{i}"] = TransformerEncoderBlock(
                input_embed_dim=self.hparams.input_embed_dim,
                num_heads=self.hparams.num_heads,
                ff_hidden_multiplier=self.hparams.ff_hidden_multiplier,
                ff_activation=self.hparams.transformer_activation,
                attn_dropout=self.hparams.attn_dropout,
                ff_dropout=self.hparams.ff_dropout,
                add_norm_dropout=self.hparams.add_norm_dropout,
                keep_attn=False,  # No easy way to convert TabTransformer Attn Weights to Feature Importance
            )
        self.transformer_blocks = nn.Sequential(self.transformer_blocks)
        self.attention_weights = [None] * self.hparams.num_attn_blocks
        if self.hparams.batch_norm_continuous_input:
            self.normalizing_batch_norm = BatchNorm1d(self.hparams.continuous_dim, self.hparams.virtual_batch_size)

        self.output_dim = self.hparams.input_embed_dim * self.hparams.categorical_dim + self.hparams.continuous_dim

    def _build_embedding_layer(self):
        return Embedding2dLayer(
            continuous_dim=0,  # Only passing and embedding categorical features
            categorical_cardinality=self.hparams.categorical_cardinality,
            embedding_dim=self.hparams.input_embed_dim,
            shared_embedding_strategy=self.hparams.share_embedding_strategy,
            frac_shared_embed=self.hparams.shared_embedding_fraction,
            embedding_bias=self.hparams.embedding_bias,
            embedding_dropout=self.hparams.embedding_dropout,
            initialization=self.hparams.embedding_initialization,
            virtual_batch_size=self.hparams.virtual_batch_size,
        )

    def forward(self, x_cat: torch.Tensor, x_cont: torch.Tensor) -> torch.Tensor:
        # (B, N)
        x = None
        if self.hparams.categorical_dim > 0:
            for i, block in enumerate(self.transformer_blocks):
                x_cat = block(x_cat)
            # Flatten (Batch, N_Categorical, Hidden) --> (Batch, N_CategoricalxHidden)
            x = rearrange(x_cat, "b n h -> b (n h)")
        if self.hparams.continuous_dim > 0:
            if self.hparams.batch_norm_continuous_input:
                x_cont = self.normalizing_batch_norm(x_cont)
            else:
                x_cont = x_cont
            # (B, N, E)
            x = x_cont if x is None else torch.cat([x, x_cont], 1)
        return x


class TabTransformerModel(BaseModel):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__(config, **kwargs)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head

    def _build_network(self):
        # Backbone
        self._backbone = TabTransformerBackbone(self.hparams)
        # Embedding Layer
        self._embedding_layer = self._backbone._build_embedding_layer()
        # Head
        self._head = self._get_head_from_config()

    # Redefining forward because this model flow is slightly different
    def forward(self, x: Dict):
        if self.hparams.categorical_dim > 0:
            x_cat = self.embed_input({"categorical": x["categorical"]})
        else:
            x_cat = None
        x = self.compute_backbone({"categorical": x_cat, "continuous": x["continuous"]})
        return self.compute_head(x)

    # Redefining compute_backbone because this model flow is slightly different
    def compute_backbone(self, x: Dict):
        # Returns output
        x = self.backbone(x["categorical"], x["continuous"])
        return x



# Pytorch Tabular
# Author: Manu Joseph <manujoseph@gmail.com>
# For license information, see LICENSE.TXT
"""DANet Model."""

import warnings

import torch
import torch.nn as nn
from omegaconf import DictConfig

from torchkeras.tabular.models.common.layers.embeddings import Embedding1dLayer

from ..base_model import BaseModel
from .arch_blocks import BasicBlock


class DANetBackbone(nn.Module):
    def __init__(
        self,
        n_continuous_features: int,
        cat_embedding_dims: list,
        n_layers: int,
        abstlay_dim_1: int,
        abstlay_dim_2: int,
        k: int,
        dropout_rate: float,
        block_activation: nn.Module,
        virtual_batch_size: int,
        embedding_dropout: float,
        batch_norm_continuous_input: bool,
    ):
        super().__init__()
        self.cat_embedding_dims = cat_embedding_dims
        self.n_continuous_features = n_continuous_features
        self.input_dim = n_continuous_features + sum([y for x, y in cat_embedding_dims])
        self.n_layers = n_layers
        self.abstlay_dim_1 = abstlay_dim_1
        self.abstlay_dim_2 = abstlay_dim_2
        self.k = k
        self.dropout_rate = dropout_rate
        self.block_activation = block_activation
        self.virtual_batch_size = virtual_batch_size
        self.batch_norm_continuous_input = batch_norm_continuous_input
        self.embedding_dropout = embedding_dropout

        self.output_dim = self.abstlay_dim_2
        self._build_network()

    def _build_network(self):
        params = {
            "fix_input_dim": self.input_dim,
            "k": self.k,
            "virtual_batch_size": self.virtual_batch_size,
            "abstlay_dim_1": self.abstlay_dim_1,
            "abstlay_dim_2": self.abstlay_dim_2,
            "drop_rate": self.dropout_rate,
            "block_activation": self.block_activation,
        }
        self.init_layer = BasicBlock(self.input_dim, **params)
        self.layers = nn.ModuleList()
        for i in range(self.n_layers - 1):
            self.layers.append(BasicBlock(self.abstlay_dim_2, **params))

    def _build_embedding_layer(self):
        return Embedding1dLayer(
            continuous_dim=self.n_continuous_features,
            categorical_embedding_dims=self.cat_embedding_dims,
            embedding_dropout=self.embedding_dropout,
            batch_norm_continuous_input=self.batch_norm_continuous_input,
            virtual_batch_size=self.virtual_batch_size,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.init_layer(x)
        for layer in self.layers:
            out = layer(x, pre_out=out)
        return out

    # Not Tested Properly
    # def _calculate_feature_importance(self):
    #     n, h, f, _ = self.attention_weights_[0].shape
    #     device = self.attention_weights_[0].device
    #     L = len(self.attention_weights_)
    #     self.local_feature_importance = torch.zeros((n, f), device=device)
    #     for attn_weights in self.attention_weights_:
    #         self.local_feature_importance += attn_weights[:, :, :, -1].sum(dim=1)
    #     self.local_feature_importance = (1 / (h * L)) * self.local_feature_importance[
    #         :, :-1
    #     ]
    #     self.feature_importance_ = (
    #         self.local_feature_importance.mean(dim=0).detach().cpu().numpy()
    #     )
    # self.feature_importance_count_+=attn_weights.shape[0]


class DANetModel(BaseModel):
    def __init__(self, config: DictConfig, **kwargs):
        super().__init__(config, **kwargs)

    @property
    def backbone(self):
        return self._backbone

    @property
    def embedding_layer(self):
        return self._embedding_layer

    @property
    def head(self):
        return self._head

    def _build_network(self):
        if self.hparams.virtual_batch_size > self.hparams.batch_size:
            warnings.warn(
                f"virtual_batch_size({self.hparams.virtual_batch_size}) is greater "
                f"than batch_size ({self.hparams.batch_size}). Setting virtual_batch_size "
                f"to {self.hparams.batch_size}. DANet uses Ghost Batch Normalization, "
                f"which works best when virtual_batch_size is small. Consider setting "
                "virtual_batch_size to something like 256 or 512."
            )
            self.hparams.virtual_batch_size = self.hparams.batch_size
        # Backbone
        self._backbone = DANetBackbone(
            cat_embedding_dims=self.hparams.embedding_dims,
            n_continuous_features=self.hparams.continuous_dim,
            n_layers=self.hparams.n_layers,
            abstlay_dim_1=self.hparams.abstlay_dim_1,
            abstlay_dim_2=self.hparams.abstlay_dim_2,
            k=self.hparams.k,
            dropout_rate=self.hparams.dropout_rate,
            block_activation=getattr(nn, self.hparams.block_activation)(),
            virtual_batch_size=self.hparams.virtual_batch_size,
            embedding_dropout=self.hparams.embedding_dropout,
            batch_norm_continuous_input=self.hparams.batch_norm_continuous_input,
        )
        # Embedding Layer
        self._embedding_layer = self._backbone._build_embedding_layer()
        # Head
        self._head = self._get_head_from_config()
